using namespace std;
/*************************************************************** 
* 
* DalColumn.cc               & XMM Science Analysis System (SAS) 
* 
* Class representing DAL Column.
* 
* (c) 1997 ESA - Written by Mark Thomas, CARA
* 
* $Id: DalColumn.cc,v 1.1.1.1 2002/08/30 13:06:29 gvacanti Exp $
* 
*****************************************************************/ 


#include <Tracer.h>
#include <SimpleScanner.h>
#include <stringutil.h>
#include <errstr.h>


#include "DalColumn.h"
#include "DalVColumnData.h"


DalColumn::DalColumn(
		const string &			name,
		DataType			dataType,
		const longVec&			size,
		const string &			label,
		const string &			units,
		const DalTable *		parent )
	:
		DalAttributable( name, label, parent -> mode(), parent ),
		DalDataComponent( dataType, size, units, this ),
		DalChild<Table>( parent ),
		DalChild<DalTable>( parent ),
		_cellType( size == zero ? Variable : Fixed ),
		_columnNumber( 0 )
{
	PROC
	ENDPROC
}


DalColumn::DalColumn(
	const DalColumn&	column,
	DalTable *		parent,
	string			newName)
	:
		DalAttributable(
			(newName.empty() ? column.name() : newName ),
			column.label(),
			column.mode(),
			column.DalChild<DalTable>::parent() ),
		DalDataComponent( column.dataType(), column.size(), column.units(), this ),
		DalChild<Table>( parent ),
		DalChild<DalTable>( parent ),
		_cellType( column.cellType() ),
		_columnNumber( 0 )
{
	PROC
	ENDPROC
}


DalColumn&
DalColumn::operator=( const Column& column )
{
	PROC

	(DalDataComponent&)*this = column;
	(DalAttributable&)*this = column;

	RETURN *this;
}


ColumnData *
DalColumn::data(
		unsigned long			from,
		unsigned long			count,
		AccessMode			accessMode) const
{
	PROC

	// *********************************************
	// Check that the column is not variable length.
	// *********************************************
	if( cellType() == Variable )
	{
		errstr << "It is not possible to call the data() function on a variable length column: " << name() << Sas::error( "invalidColumnOperation" );
	} // if

	if( accessMode == AsParent )
	{
		accessMode = this->DalChild<Table>::parent()->parent()->mode();
	} // if

	ColumnData * newColumnData = makeColumnData(
				from,
				count,
				accessMode );

	_dataObjects.push_back( newColumnData );

	RETURN newColumnData;

}


ColumnData *
DalColumn::rowData(
		unsigned long			rowNumber,
		AccessMode			accessMode ) const
{
	PROC

	// ***********************************
	// Check that the row number is valid.
	// ***********************************
	const unsigned long rows = DalChild<Table>::parent() -> rows();

	if( rowNumber >= rows )
	{
		errstr << "The specified row number " << rowNumber << " is invalid for column: " << name() << ", which has " << rows << " rows. The row number must be in the range 0 to " << rows - 1 << "." << Sas::error( "invalidRowNumber " );
	} // if


	if( accessMode == AsParent )
	{
		accessMode = this->DalChild<Table>::parent()->parent()->mode();
	} // if

	ColumnData* newColumnData = 0;

	// *********************************************
	// Check that the column is not variable length.
	// *********************************************
	if( cellType() == Variable )
	{

		CellData* newCellData = makeCellData(
						rowNumber,
						0,
						0,
						accessMode );

		_dataObjects.push_back( newCellData );

		newColumnData = new DalVColumnData( newCellData );

	}
	else
	{
		newColumnData = makeColumnData(
						rowNumber,
						1,
						accessMode );

		_dataObjects.push_back( newColumnData );
	} // if


	RETURN newColumnData;

}



CellData *
DalColumn::cellData(
		unsigned long			rowNumber,
		unsigned long			from,
		unsigned long			count,
		AccessMode			accessMode ) const
{
	PROC

	// ***********************************
	// Check that the row number is valid.
	// ***********************************
	const unsigned long rows = DalChild<Table>::parent() -> rows();

	if( rowNumber >= rows )
	{
		errstr << "The specified row number " << rowNumber << " is invalid for column: " << name() << ", which has " << rows << " rows. The row number must be in the range 0 to " << rows - 1 << "." << Sas::error( "invalidRowNumber " );
	} // if


	// ******************************************
	// Check that the column is not fixed length.
	// ******************************************
	if( cellType() == Fixed )
	{
		errstr << "The cellData() function may not be called on the fixed length column: " << name() << Sas::error( "invalidColumnOperation" );

	} // if

	if( accessMode == AsParent )
	{
		accessMode = this->DalChild<Table>::parent()->parent()->mode();
	} // if

	CellData* newCellData = makeCellData(
					rowNumber,
					from,
					count,
					accessMode );

	_dataObjects.push_back( newCellData );

	RETURN newCellData;
}


string
DalColumn::parseName( const string & compoundName ) const
{
	PROC

	SimpleScanner	sc( compoundName, ":%" );
	string		parsedName;

	try
	{
		parsedName = sc.match( SimpleToken::String ).asString();

		if( sc.nextIsCharacter( ':' ) )
		{
			const DalTable * parentTable = ((DalChild<DalTable>&)*this).parent();
			parsedName = parentTable -> parseName( compoundName );
			SimpleScanner	sc( parsedName, ":%" );

			if( sc.nextIsCharacter( ':' ) )
			{
				sc.skip();

				ALTRETURNV( sc.remainder() );
			} // if

			parsedName = sc.match( SimpleToken::String ).asString();

			if( parsedName != name() )
			{
				errstr << "The compound name '" << compoundName << "' is invalid; the qualified name of this object is " << qualifiedName() << Sas::error( "invalidCompoundName" );
			} // if
			if( sc.nextIsCharacter( '%' ) )
			{
				ALTRETURNV( sc.remainder() );
			} // if
			sc.match( ':' );
			ALTRETURNV( sc.remainder() );
		}
		RETURN compoundName;
	}
	catch( SimpleScanner::Expected e )
	{
		errstr << "An error was detected with the name '" << compoundName << "'.  " << e.message() << Sas::error( "invalidCompoundName" );
	}

	RETURN "will not get here";

}


string
DalColumn::qualifiedName() const
{
	PROC
	RETURN DalChild<Table>::parent() -> qualifiedName() + ":" + name();
}


unsigned int
DalColumn::columnNumber() const
{
	PROC
	RETURN _columnNumber;
}


void
DalColumn::columnNumber( unsigned int number )
{
	PROC
	_columnNumber = number;
	ENDPROC
}


Column::CellType
DalColumn::cellType() const
{
	PROC
	RETURN _cellType;
}


