/*************************************************************** 
* 
* DalData.h          & XMM Science Analysis System (SAS) 
* 
* Class representing Data
* 
* (c) 1997 ESA - Written by Mark Thomas, CARA
* 
* $Id: DalData.h,v 1.2 2002/09/06 14:57:37 gvacanti Exp $
* 
*****************************************************************/ 
#ifndef DAL_DATA_H
#define DAL_DATA_H


#include <errstr.h>
#include <Tracer.h>
#include <NAN.h>


#include "Dal.h"
#include "DalChild.h"
#include "DalSeekable.h"
#include "DalDal.h"


template<class T>
class DalData : public virtual Data, public DalDal, public DalChild<T>
{
	public :

		DalData(
			const T *	parent,
			unsigned long	from,
			unsigned long	count,
			AccessMode	accessMode );
		virtual ~DalData() {}

		void setToNull( unsigned long pos ) const;
		bool isNull( unsigned long pos ) const;
		bool hasNulls() const;

		DalSeekable<unsigned long> & seekable() const { PROC RETURN const_cast< DalSeekable<unsigned long> & >( _seekable ); }

		virtual void release() {}

	private :
		DalSeekable<unsigned long>	_seekable;

};


template<class T>
inline DalData<T>::DalData(
	const T *	parent,
	unsigned long	from,
	unsigned long	count,
	AccessMode	accessMode )
	:
		DalDal( accessMode, parent ),
		DalChild<T>( parent ),
		_seekable( from, count )
{
	PROC
	ENDPROC
}


template<class T>
inline void
DalData<T>::setToNull( unsigned long pos ) const
{
	PROC
	switch( dataType() )
	{
		case Data::Bool :

			errstr << "Boolean Null Values are not supported." << Sas::error( "notSupported" );

			break;
		case Data::Int8 :
		{
			int8 * i8 = int8Data();
			i8[ pos ] = parent() -> intNullValue();
		}
			break;
		case Data::Int16 :
		{
			int16 * i16 = int16Data();
			i16[ pos ] = parent() -> intNullValue();
		}
			break;
		case Data::Int32 :
		{
			int32 * i32 = int32Data();
			i32[ pos ] = parent() -> intNullValue();
		}
			break;
		case Data::Real32 :
		{
			real32 * r32 = real32Data();
			NaN.set( r32[ pos ] );
			//r32[ pos ] = parent() -> realNullValue();
		}
			break;
		case Data::Real64 :
		{
			real64 * r64 = real64Data();
			NaN.set( r64[ pos ] );
			//r64[ pos ] = parent() -> realNullValue();
		}
			break;
		case Data::DString :

			errstr << "String Null Values are not supported." << Sas::error( "notSupported" );

			break;
		default :
			errstr << "Invalid data type." << Sas::error( "internalError" );
			break;
	} // switch
	ENDPROC
}


template<class T>
inline bool
DalData<T>::isNull( unsigned long pos ) const
{
	PROC
	bool	isnull = false;

	switch( dataType() )
	{
		case Data::Bool :

			errstr << "Boolean Null Values are not supported." << Sas::error( "notSupported" );

			break;
		case Data::Int8 :
		{
			int8 * i8 = int8Data();
			if( i8[ pos ] == parent() -> intNullValue() ) isnull = true;
		}
			break;
		case Data::Int16 :
		{
			int16 * i16 = int16Data();
			if( i16[ pos ] == parent() -> intNullValue() ) isnull = true;
		}
			break;
		case Data::Int32 :
		{
			int32 * i32 = int32Data();
			if( i32[ pos ] == parent() -> intNullValue() ) isnull = true;
		}
			break;
		case Data::Real32 :
		{
			real32 * r32 = real32Data();
			isnull = NaN.isSet( r32[ pos ] );
			//if( r32[ pos ] == parent() -> realNullValue() ) isnull = true;
		}
			break;
		case Data::Real64 :
		{
			real64 * r64 = real64Data();
			isnull = NaN.isSet( r64[ pos ] );
			//if( r64[ pos ] == parent() -> realNullValue() ) isnull = true;
		}
			break;
		case Data::DString :

			errstr << "String Null Values are not supported." << Sas::error( "notSupported" );

			break;
		default :
			errstr << "Invalid data type." << Sas::error( "internalError" );
			break;
	} // switch
	RETURN isnull;
}


template<class T>
inline bool
DalData<T>::hasNulls() const
{
	PROC

	bool	hasnulls = false;

	for( unsigned long i = 0; i < _seekable.count(); ++i )
	{
		if( hasnulls = isNull( i ) ) break;
	} // for

	RETURN hasnulls;

}

#endif




