using namespace std;
/*************************************************************** 
* 
* DalNullable.cc          & XMM Science Analysis System (SAS) 
* 
* Implementation of class representing a null value.
* 
* (c) 1997 ESA - Written by Mark Thomas, CARA
* 
* $Id: DalNullable.cc,v 1.1.1.1 2002/08/30 13:06:29 gvacanti Exp $
* 
*****************************************************************/ 

#include <Tracer.h>
#include <errstr.h>


#include "DalNullable.h"


class Nvalue
{
	public :
		Nvalue() : _defined( false ) {}
		virtual ~Nvalue() {}
		bool nullDefined() const { PROC RETURN _defined; }
		virtual Nullable::NullType nullType() const { PROC RETURN Nullable::Undefined; }
		virtual void nullValue( long ) { operationNotAllowed(); }
		virtual void nullValue( double ) { operationNotAllowed(); }
		virtual void nullValue( const string& ) { operationNotAllowed(); }
		virtual long intNullValue() const { operationNotAllowed(); return 0; }
		virtual double realNullValue() const { operationNotAllowed(); return 0.0; }
		virtual const string& stringNullValue() const { operationNotAllowed(); static string s; return s; }
		//void deleteNullValue() { PROC _defined = false; ENDPROC }
		virtual void deleteNullValue() { operationNotAllowed(); }
		void checkIfDefined() const
		{
			PROC
			if( !nullDefined() ) errstr << "The null value has not been defined, and the operation is thus invalid" << Sas::error( "nullNotDefined" );
			ENDPROC
		}
	protected :
		bool _defined;
	private :

		void operationNotAllowed() const
		{
			errstr << "The requested null value operation is not allowed." << Sas::error( "invalidNullOperation" );
		}
};


class IntNvalue : public Nvalue
{
	public :
		Nullable::NullType nullType() const { PROC RETURN Nullable::Integer; }
		void nullValue( long value ) { PROC _value = value; _defined = true; ENDPROC }
		long intNullValue() const { PROC checkIfDefined(); RETURN _value; }
		void deleteNullValue() { PROC _defined = false; ENDPROC }
	private :
		long _value;
};


class RealNvalue : public Nvalue
{
	public :
		RealNvalue() { PROC _defined = true; ENDPROC }
		Nullable::NullType nullType() const { PROC RETURN Nullable::Real; }
		void deleteNullValue() { PROC ENDPROC }
	private :
		//double _value;
};


class StringNvalue : public Nvalue
{
	public :
		Nullable::NullType nullType() const { PROC RETURN Nullable::String; }
		void nullValue( const string& value ) { PROC _value = value; _defined = true; ENDPROC }
		const string& stringNullValue() const { PROC checkIfDefined(); RETURN _value; }
		void deleteNullValue() { PROC _defined = false; ENDPROC }
	private :
		string _value;
};


DalNullable::DalNullable( NullType nullType )
	: _nvalue(0)
{
	PROC
	switch( nullType )
	{
		case Nullable::Integer :
			_nvalue = new IntNvalue();
			break;
		case Nullable::Real :
			_nvalue = new RealNvalue();
			break;
		case Nullable::String :
			//_nvalue = new StringNvalue();
			break;
		case Nullable::Undefined :
			break;
		default :
			errstr << "Invalid switch statement." << Sas::error( "internalError" );
			break;
	} // switch
	ENDPROC
}


DalNullable::~DalNullable()
{
	PROC
	delete _nvalue;
	ENDPROC
}


DalNullable::DalNullable( const Nullable& n )
	: _nvalue( 0 )
{
	PROC
	operator=( n );
	ENDPROC
}


DalNullable&
DalNullable::operator=( const Nullable& n )
{
	PROC
	delete _nvalue;
	_nvalue = 0;

	switch( n.nullType() )
	{
		case DalNullable::Integer :
			_nvalue = new IntNvalue();
			_nvalue -> nullValue( n.intNullValue() );
			break;
		case DalNullable::Real :
			_nvalue = new RealNvalue();
			break;
		case DalNullable::String :
			//_nvalue = new StringNvalue();
			//_nvalue -> nullValue( n.stringNullValue() );
			break;
		case DalNullable::Undefined : // fall through
			break;
		default :
			errstr << "Invalid switch statement." << Sas::error( "internalError" );
			break;
	} // switch

	RETURN *this;
}


void
DalNullable::checkOperation( Nullable::NullType nType ) const
{
	PROC
	if( nType != nullType() )
	{
		errstr << "The nullable operation is invalid." << Sas::error( "invalidNullableOperation" );
	} // if
	ENDPROC
}


bool
DalNullable::nullDefined() const
{
	PROC
	RETURN _nvalue != 0 && _nvalue -> nullDefined();
}


Nullable::NullType
DalNullable::nullType() const
{
	PROC
	RETURN _nvalue ? _nvalue -> nullType() : Nullable::Undefined;
}


void
DalNullable::nullValue( long value )
{
	PROC
	checkOperation( Nullable::Integer );
	_nvalue -> nullValue( value );
	ENDPROC
}


void
DalNullable::nullValue( const string& value )
{
	PROC
	checkOperation( Nullable::String );
	_nvalue -> nullValue( value );
	ENDPROC
}

long
DalNullable::intNullValue() const
{
	PROC
	checkOperation( Nullable::Integer );
	RETURN _nvalue -> intNullValue();
}


const string&
DalNullable::stringNullValue() const
{
	PROC
	checkOperation( Nullable::String );
	RETURN _nvalue -> stringNullValue();
}


void
DalNullable::deleteNullValue()
{
	PROC
	if( !nullDefined() )
	{
		errstr << "The null value of the object is not set and so cannot be deleted" << Sas::error( "nullValueNotDefined" );
	} // if
	_nvalue -> deleteNullValue();
	ENDPROC
}

