using namespace std;
/*************************************************************** 
 * 
 * DalTable.cc & XMM Science Analysis System (SAS) 
 * 
 * Class representing DAL Table. 
 * 
 * (c) 1997 ESA - Written by Mark Thomas, CARA
 * 
 * $Id: DalTable.cc,v 1.2 2002/09/06 14:57:37 gvacanti Exp $
 * 
 *****************************************************************/ 


#include <Tracer.h>
#include <SimpleScanner.h>
#include <stringutil.h>

#include "dal_utilities.h"
#include "DalTable.h"
#include "DalColumn.h"


DalTable::DalTable(
    const string &		name,
    const string &		label,
    unsigned long		rows,
    DalDataSet *		parent )
    :
    DalBlock( name, Block::TableT, label, parent ),
    DalSeekable<unsigned long>( 0, rows ),
    _rows( rows )
{
    PROC
	ENDPROC
	}


void
DalTable::checkValidColumnNumber( unsigned int columnNumber ) const
{
    PROC
	// ***********************************************
	// Check that the required column number is valid.
	// ***********************************************
	if( columnNumber >= _columns.size() )
	{
	    errstr << "Table " << name() << " has no column " << columnNumber << "; it has " << columns() << " columns" << Sas::error("invalidColumnNumber");

	} // if
    ENDPROC
	}


Column *
DalTable::addColumn(
    const string &			name,
    Data::DataType			dataType,
    const string &			label,
    const string &			units,
    const longVec &			size,
    int				        position)
{
    PROC

	checkIfReadOnly();

    string	parsedName = parseName( name );

    SimpleScanner	sc( parsedName, ":%" );

    string columnName;

    try
    {
	columnName = sc.match( SimpleToken::String ).asString();
    }
    catch( SimpleScanner::Expected e )
    {
	errstr << "The compound name " << name << " is invalid; the qualified name of this object is " << qualifiedName() << Sas::error( "invalidCompoundName" );
    } // if


    // ********************************************
    // Check that a column with the given name does
    // not already exist.
    // ********************************************
    if ( hasColumn( columnName ) )
	errstr << "cannot add column '" << columnName << "': table '"
	       << this->name() << "' already contains a column with this name"
	       << Sas::error("columnAlreadyExists");

    // 2002-01-17 (GV) Remove this check, as it is acceptable to
    // have a vector column with n dimensions of 0 extent.
    // ********************************************************************
    // If the column is a vector column, then check that the size is valid.
    // ********************************************************************
//  	if( size.size() && ::numberOfElements( size ) <= 0 )
//  	{
//  		errstr << "The vector column with qualified name '" << qualifiedName() << "')" << " has an invalid size since it gives rise to a non-positive number of elements" << Sas::error( "invalidVectorColumnSize" );
//  	} // if

    DalColumn *	newColumn = makeColumn(
	columnName,
	dataType,
	label,
	units,
	size,
	this );


    insert( newColumn, position );

    // ********************************
    // Add an entry into the data file.
    // ********************************
    DataSetReaderWriter * dataSetReaderWriter = DalChild<DalDataSet>::parent() -> dataSetReaderWriter();
    if( dataSetReaderWriter ) dataSetReaderWriter -> addEntry( *newColumn );

    RETURN newColumn;
}


Column *
DalTable::add(
    const Column *			column,
    const string &			newName,
    int				        position)
{
    PROC
 
	const string	columnName = newName.empty() ? column -> name() : newName;

    if ( hasColumn( columnName ))
    {
	errstr << "cannot add column '" << newName << "': table '" << name() << " already contains a column with this name" << Sas::error("columnExists");
    } // if

    DalColumn *	newColumn = makeColumn(
	columnName,
	column -> dataType(),
	column -> label(),
	column -> units(),
	column -> size(),
	this );

    insert( newColumn, position );

    // ********************************
    // Add an entry into the data file.
    // ********************************
    DataSetReaderWriter * dataSetReaderWriter = DalChild<DalDataSet>::parent() -> dataSetReaderWriter();
    if( dataSetReaderWriter ) dataSetReaderWriter -> addEntry( *newColumn );

    *newColumn = *column;

    RETURN newColumn;
}


unsigned long
DalTable::rows() const
{
    PROC
	RETURN _rows;
}


Column *
DalTable::column(
    unsigned int	columnNumber,
    AccessMode	accessMode )
{

    PROC

	checkValidColumnNumber( columnNumber );

    DalColumn * col = _columns[ columnNumber ];

    col -> mode( accessMode );
    col -> alwaysAllowModify( false );

    RETURN col;
}


const Column *
DalTable::column(
    unsigned int	columnNumber,
    AccessMode	accessMode ) const
{
    PROC

	checkValidColumnNumber( columnNumber );

    DalColumn * col = _columns[ columnNumber ];

    col -> mode( accessMode );
    col -> alwaysAllowModify( false );

    RETURN col;
}


Column *
DalTable::column(
    const string &	columnName,
    AccessMode	accessMode )
{

    PROC
	RETURN column( columnNumber( columnName ), accessMode );
}


const Column *
DalTable::column(
    const string &	columnName,
    AccessMode	accessMode ) const
{

    PROC
	RETURN column( columnNumber( columnName ), accessMode );
}


bool
DalTable::hasColumn( const string & name ) const
{
    PROC

	string	parsedName = parseName( name );

    SimpleScanner	sc( parsedName, ":%" );

    string	columnName;

    try
    {
	columnName = sc.match( SimpleToken::String ).asString();
    }
    catch( SimpleScanner::Expected e )
    {
	errstr << "The compound name " << name << " is invalid; the qualified name of this object is " << qualifiedName() << Sas::error( "invalidCompoundName" );
    } // if

    for(
	ColumnList::const_iterator i = _columns.begin();
	i != _columns.end();
	++i )
    {
	if( columnName == (*i) -> name() )
	{
	    ALTRETURNV( true );

	    break;

	} // if
    } // for

    RETURN false;
}


void
DalTable::deleteColumn( const string & columnName )
{

    PROC
	deleteColumn( columnNumber( columnName ));
    ENDPROC
	}


void
DalTable::deleteColumn( unsigned int columnNumber )
{
    PROC
    
	checkIfReadOnly();
    checkValidColumnNumber( columnNumber );
  
    ColumnList::iterator	i;
    unsigned int		j = 0;
  
  
    for( i = _columns.begin(); i != _columns.end(); ++i )
    {
	if( j == columnNumber )
	{
	    // ********************************************
	    // Before deleting the column from the table, we need to flush
	    // the column's data
	    // ********************************************
	    (*i)->release();
	    // ************************************
	    // Delete the entry from the data file.
	    // ************************************
	    DalChild<DalDataSet>::parent() -> dataSetReaderWriter() -> deleteEntry( *(*i) );
	  
	    // *********************************
	    // Delete the column from the table.
	    // *********************************
	    delete *i;
	    _columns.erase( i );
	  
	    // *****************
	    // Renumber columns.
	    // *****************
	    i = _columns.begin();
	    for( unsigned int j = 0; j < _columns.size(); ++j, ++i )
	    {
		(*i) -> columnNumber( j );
	    } // for
	  
	    ALTRETURN;
	  
	} // if
      
	j++;
      
    } // for
  
    errstr << "there is no column " << columnNumber << " to delete from table " << name() << "; it has " << columns() << " columns" << Sas::error("invalidColumnNumber");
  
    ENDPROC;
  
}


unsigned int
DalTable::columns() const
{
    PROC
	RETURN _columns.size();

}


unsigned int
DalTable::columnNumber( const string & columnName ) const
{

    PROC

	string	parsedName = parseName( columnName );

    SimpleScanner	sc( parsedName, ":%" );
    unsigned int	columnNumber = 0;

    if( sc.nextIsCharacter( ':' ) )
    {
	errstr << "The compound name " << columnName << " is invalid; the qualified name of this object is " << qualifiedName() << Sas::error( "invalidCompoundName" );
    }
    else if( !sc.nextIsCharacter( '%' ) && !sc.eof() )
    {
	try
	{
	    parsedName = sc.match( SimpleToken::String ).asString();
	}
	catch( SimpleScanner::Expected e )
	{
	    errstr << "The compound name " << columnName << " is invalid; the qualified name of this object is " << qualifiedName() << Sas::error( "invalidCompoundName" );
	} // if

	ColumnList::const_iterator i;

	for( i = _columns.begin(); i != _columns.end(); ++i, ++columnNumber )
	{
	    SAS_DEBUG( clog << "encountered column " << (*i)->name() << endl; )

		if( (*i)->name() == parsedName )
		{
		    break;

		} // if
	} // for

	if( i == _columns.end() )
	{
	    DataSet * dataSet = ((DalChild<DataSet>&)*this).parent();

	    errstr << "Could not find column with qualified name " << qualifiedName() << ":" << parsedName << " in table " << name() << " in dataset " << dataSet -> name() << Sas::error("columnNotFound");

	} // if
    } // if

    RETURN columnNumber;
}


void
DalTable::copyRows( unsigned long from, unsigned long to, unsigned long count )
{
    PROC

    if(from != to){
	    
	checkIfReadOnly();
	
	for(
	    ColumnList::iterator i = _columns.begin();
	    i != _columns.end();
	    ++i )
	{
	    (*i)->copyRows( from, to, count );
	    
	} // for	
    }
    ENDPROC
}


void
DalTable::deleteRows( unsigned long from, unsigned long count )
{
    PROC

	checkIfReadOnly();

    if( from >= _rows ) ALTRETURN;

    count = min( _rows - from, count );

    //releaseAll();
    release();

    // ***************************
    // Delete rows into data file.
    // ***************************
    DalChild<DalDataSet>::parent() -> dataSetReaderWriter() -> deleteRows( *this, from, count );

    _rows -= count;

    DalSeekable<unsigned long>::seek( 0, _rows );

    for(
	ColumnList::iterator i = _columns.begin();
	i != _columns.end();
	++i )
    {
	(*i)->deleteRows( from, count );
    } // for

    //_rows -= count;

    //if (count < _rows)
    //	_rows -= count;
    //else
    //	_rows = 0;

    ENDPROC
	}


void
DalTable::insertRows( unsigned long pos, unsigned long count )
{
    PROC

	checkIfReadOnly();

    if( pos >= _rows ) pos = _rows;

    //releaseAll();
    release();

    // ***************************
    // Insert rows into data file.
    // ***************************
    DalChild<DalDataSet>::parent() -> dataSetReaderWriter() -> insertRows( *this, pos, count );

    _rows += count;

    DalSeekable<unsigned long>::seek( 0, _rows );

    for(
	ColumnList::iterator i = _columns.begin();
	i != _columns.end();
	++i )
    {
	(*i)->insertRows( pos, count );
    } // for

    //_rows += count;

    ENDPROC
	}


void
DalTable::insert( DalColumn * column, int position )
{
    PROC

	// ***************************************************
	// The intention is to place the new item in the list
	// so that, in the new modified
	// list, it has position Position.
	// ***************************************************

	// ********************************************
	// Check that position is within a valid range.
	// position is unsigned and > 0.
	// ********************************************
	unsigned int	pos = (position == -1 ) ? _columns.size() : position;

    if( pos > _columns.size() )
    {
	errstr << "cannot add the column " << column->name() << " at position " << pos << " in table " << this->name() << "; it has " << columns() << " columns" << Sas::error("invalidColumnPosition");
    } // if

    ColumnList::iterator	i = _columns.begin();

    for( unsigned int j = 0; j < pos; ++j ) ++i;

    _columns.insert( i, column );

    // *****************
    // Renumber columns.
    // *****************
    i = _columns.begin();
    for( unsigned int j = 0; j < _columns.size(); ++j, ++i )
    {
	(*i) -> columnNumber( j );
    } // for

    ENDPROC
	}


//void
//DalTable::releaseAll()
//{
//	PROC
//	for(
//		ColumnList::iterator i = _columns.begin();
//		i != _columns.end();
//		++i )
//	{
//		(*i)->releaseAll();
//	} // for
//	ENDPROC
//}


void
DalTable::printOn( ostream & os ) const
{
    PROC

	// ****************
	// Write the Block.
	// ****************
	DalBlock::printOn( os );

    // ******************
    // Write the columns.
    // ******************

    os << "There are " << _columns.size() << " Columns" << endl;

    unsigned int no = 1;

    for(
	ColumnList::const_iterator i = _columns.begin();
	i != _columns.end();
	++i, ++no )
    {
	os << "[" << no << "]";

	(*i) -> printOn( os );

    } // for

    ENDPROC
	}


void
DalTable::release()
{
    PROC
	for( unsigned int i = 0; i < _columns.size(); ++i )
	{
	    _columns[ i ] -> DalDataComponent::release();
	    // This makes it possible to add rows to a table that
	    // has been released.
	    ////		_columns[i]->alwaysAllowModify(true);
	    // 2001-10-11 MT suggests the following instead.
	    this->column(i);
	} // for
    ENDPROC
	}


void
DalTable::seek(
    unsigned long		from,
    unsigned long		count )
{
    PROC
	for( unsigned int i = 0; i < _columns.size(); ++i )
	{
	    _columns[ i ] -> seek( from, count );
	} // for
    DalSeekable<unsigned long>::seek( from, count );
    ENDPROC
	}


