using namespace std;
/*************************************************************** 
 * 
 * HiLowColumn.cc               & XMM Science Analysis System (SAS) 
 * 
 * Class representing Table Column.
 * 
 * (c) 1997 ESA - Written by Mark Thomas, CARA
 * 
 * $Id: HiLowColumn.cc,v 1.2 2002/09/06 14:57:37 gvacanti Exp $
 * 
 *****************************************************************/ 


#include <errstr.h>
#include <Tracer.h>

#include "HiLowColumn.h"
#include "HiLowColumnData.h"
#include "HiLowCellData.h"
#include "machine_specific_types.h"
#include "Storage.h"
#include "HiLowCellData.h"


typedef Storage<HiLowColumnData> ColumnDataStorage;
typedef Storage<HiLowCellData> CellDataStorage;


static unsigned long
numberOfElements( const longVec & size )
{
    PROC

	unsigned long elements = 0;

    if( !( size == Dal::zero ) )
    {
	elements = 1;
	for( unsigned int i = 0; i < size.size(); ++i )
	    elements *= size[ i ];
    } // if

    RETURN elements;
}


HiLowColumn::HiLowColumn(
    const string &			name,
    DataType			dataType,
    const string &			label,
    const string &			units,
    const longVec &			size,
    const DalTable *		parent )

    :
    DalColumn( name, dataType, size, label, units, parent ),
    DalSeekable<unsigned long>(
	parent -> from(),
	parent -> count() )
{

    PROC

	errstr << "Constructing HiLowColumn with qualified name='"
	       << qualifiedName() <<"' dataType="<<int( dataType )
	       << " rows="<<parent->rows()<<" dimensions="
	       << dimensions() <<" units="<<units
	       << Sas::message( Msg::Core,Msg::Verbose );

    if( cellType() == Fixed )
    {
	_data = new (HiLowDataStorage *)[ 1 ];
	_data[ 0 ] = new ColumnDataStorage(
	    parent -> name() + "." + name,
	    dataType,
	    parent -> rows() * ::numberOfElements( size ),
	    0 );
    }
    else
    {
	_data = new (HiLowDataStorage *)[ parent -> rows() ];
	for( unsigned long i = 0; i < parent -> rows(); ++i )
	{
	    _data[ i ] = new CellDataStorage(
		parent -> name() + "." + name,
		dataType,
		0,
		0 );

	} // for
    } // if

    ENDPROC

	}


HiLowColumn::~HiLowColumn()
{
    PROC

	// ***********************************
	// Release and delete the column data. 
	// ***********************************
	//release();

	checkUnreleasedData();

    //while( _dataObjects.size() )
    //{
    //	errstr << "Unreleased data detected for column with qualified name " << qualifiedName() << Sas::warning( "unreleasedData" );
    //	delete _dataObjects[ _dataObjects.size() - 1 ];
    //} // while

    if( cellType() == Fixed )
    {
	delete _data[ 0 ];
    }
    else
    {
	for( unsigned int i = 0; i < DalChild<Table>::parent() -> rows(); ++i )
	{
	    delete _data[ i ];
	} // for
    } // if
    delete[] _data;
    ENDPROC
	}


HiLowColumn::HiLowColumn(
    const DalColumn &	column,
    DalTable * 		parent,
    string 			newName)
    :
    DalColumn( column, parent, newName ),
    //DalColumn(
    //	newName.empty() ? column.name() : newName,
    //	column.dataType(),
    //	column.size(),
    //	column.units(),
    //	column.label(),
    //	parent ),
    DalSeekable<unsigned long>( parent -> from(), parent -> count() )
{
    PROC

	if( column.cellType() == Fixed )
	{
	    _data = new (HiLowDataStorage *)[ 1 ];
	    _data[ 0 ] = new ColumnDataStorage(
		name()+"."+newName,
		column.dataType(),
		parent -> rows() * ::numberOfElements( column.size()),
		0 );
	}
	else
	{
	    _data = new (HiLowDataStorage *)[ parent -> rows() ];
	    for( unsigned long i = 0; i < parent -> rows(); ++i )
	    {
		_data[ i ] = new CellDataStorage(
		    name()+"."+newName,
		    dataType(),
		    0,
		    0 );

	    } // for
	} // if

    operator=(column);

    ENDPROC

	}


void
HiLowColumn::assignFixed( const Column & column )
{
    PROC

	delete _data[ 0 ];
    _data[ 0 ] = new ColumnDataStorage(
	name(),
	column.dataType(),
	rows() * ::numberOfElements( size() ),
	0 );

    ColumnData *s = column.data();
    ColumnData *d = data();
    unsigned long N = rows() * elements();

    switch( dataType() )
    {
	case Bool :
	{
	    bool8 *sd = s->boolData();
	    bool8 *dd = d->boolData();
	    for (unsigned long i = 0; i < N; ++i)
		dd[i] = sd[i];
	}
	break;

	case Int8:
	{
	    int8 *sd = s->int8Data();
	    int8 *dd = d->int8Data();
	    for (unsigned long i = 0; i < N; ++i)
		dd[i] = sd[i];
	}
	break;

	case Uint16:
	{
	    uint16 *sd = s->uint16Data();
	    uint16 *dd = d->uint16Data();
	    for (unsigned long i = 0; i < N; ++i)
		dd[i] = sd[i];
	}
	break;

	case Int16:
	{
	    int16 *sd = s->int16Data();
	    int16 *dd = d->int16Data();
	    for (unsigned long i = 0; i < N; ++i)
		dd[i] = sd[i];
	}
	break;

	case Uint32:
	{
	    uint32 *sd = s->uint32Data();
	    uint32 *dd = d->uint32Data();
	    for (unsigned long i = 0; i < N; ++i)
		dd[i] = sd[i];
	}
	break;

	case Int32:
	{
	    int32 *sd = s->int32Data();
	    int32 *dd = d->int32Data();
	    for (unsigned long i = 0; i < N; ++i)
		dd[i] = sd[i];
	}
	break;

	case Real32:
	{
	    real32 *sd = s->real32Data();
	    real32 *dd = d->real32Data();
	    for (unsigned long i = 0; i < N; ++i)
		dd[i] = sd[i];
	}
	break;
	case Real64:
	{
	    real64 *sd = s->real64Data();
	    real64 *dd = d->real64Data();
	    for (unsigned long i = 0; i < N; ++i)
		dd[i] = sd[i];
	}
	break;

	case DString:
	{
	    char * sd = s->stringData();
	    char * dd = d->stringData();
	    for (unsigned long i = 0; i < N; ++i)
		dd[i] = sd[i];
	}
	break;

	default:
	    errstr << "invalid type " << int( dataType() ) << " in column " << column.name() << Sas::error("internalError");
    } // switch

    delete d;
    delete s;
    ENDPROC
	}


void
HiLowColumn::assignVariable( const Column & column )
{
    PROC
	for(unsigned k = 0; k < column.rows(); ++k)
	    delete _data[k];

    _data = new (HiLowDataStorage *)[column.rows()];
    for( unsigned long i = 0; i < column.rows(); ++i ){
	CellData * s = column.cellData(i);
	_data[i] = new CellDataStorage(name(),
				       column.dataType(),
				       0,
				       0);
	CellData * d = cellData(i);
	d->size(s->size());
	switch(column.dataType()) {
	    case Bool:
	    {
		bool8 * sd = s->boolData();
		bool8 * dd = d->boolData();
		for(unsigned long k = 0; k < s->size(); ++k)
		    dd[k] = sd[k];
		break;
	    }
	    case Int8:
	    {
		int8 * sd = s->int8Data();
		int8 * dd = d->int8Data();
		for(unsigned long k = 0; k < s->size(); ++k)
		    dd[k] = sd[k];
		break;
	    }
	    case Int16:
	    {
		int16 * sd = s->int16Data();
		int16 * dd = d->int16Data();
		for(unsigned long k = 0; k < s->size(); ++k)
		    dd[k] = sd[k];
		break;
	    }
	    case Uint16:
	    {
		uint16 * sd = s->uint16Data();
		uint16 * dd = d->uint16Data();
		for(unsigned long k = 0; k < s->size(); ++k)
		    dd[k] = sd[k];
		break;
	    }
	    case Int32:
	    {
		int32 * sd = s->int32Data();
		int32 * dd = d->int32Data();
		for(unsigned long k = 0; k < s->size(); ++k)
		    dd[k] = sd[k];
		break;
	    }
	    case Uint32:
	    {
		uint32 * sd = s->uint32Data();
		uint32 * dd = d->uint32Data();
		for(unsigned long k = 0; k < s->size(); ++k)
		    dd[k] = sd[k];
		break;
	    }
	    case Real32:
	    {
		real32 * sd = s->real32Data();
		real32 * dd = d->real32Data();
		for(unsigned long k = 0; k < s->size(); ++k)
		    dd[k] = sd[k];
		break;
	    }
	    case Real64:
	    {
		real64 * sd = s->real64Data();
		real64 * dd = d->real64Data();
		for(unsigned long k = 0; k < s->size(); ++k)
		    dd[k] = sd[k];
		break;
	    }
	    case DString:
	    {
		char * sd = s->stringData();
		char * dd = d->stringData();
		for(unsigned long k = 0; k < s->size(); ++k)
		    dd[k] = sd[k];
		break;
	    }
	}

	delete d;
	delete s;
    }
			       
    ENDPROC
	}


HiLowColumn &
HiLowColumn::operator=( const Column & column )
{

    PROC

	(Attributable&)*this = column;

    //relabel(column.label());

    //(DalChild<Table> &)*this = column;
    //(DalChild<HiLowTable> &)*this = column;
    //(HiLowLabelled &)*this = column;

    //_cellType = column.cellType();
    //_dataType = column.dataType();
    //*_data = *(column._data);
    //_size = column.size();

    if( cellType() == Fixed )
    {
	assignFixed( column );
    }
    else
    {
	assignVariable( column );
    } // if

    RETURN *this;

}


ColumnData *
HiLowColumn::makeColumnData(
    unsigned long			from,
    unsigned long			count,
    AccessMode			accessMode ) const
{
    PROC
	RETURN new HiLowColumnData(
	    _data[ 0 ],
	    this,
	    from,
	    count,
	    accessMode );
}


CellData *
HiLowColumn::makeCellData(
    unsigned long			rowNumber,
    unsigned long			from,
    unsigned long			count,
    AccessMode			accessMode ) const
{
    PROC
	RETURN new HiLowCellData(
	    _data[ rowNumber ],
	    this,
	    rowNumber,
	    from,
	    count,
	    accessMode );

}

/*
  ColumnData *
  HiLowColumn::data(
  unsigned long			from=0,
  unsigned long			count=0,
  AccessMode			mode=AsParent ) const
  {
  PROC

  // *********************************************
  // Check that the column is not variable length.
  // *********************************************
  if( cellType() == Variable )
  {
  errstr << "It is not possible to call the data() function on a variable length column: " << name() << Sas::error( "invalidColumnOperation" );
  } // if

  if ( mode == AsParent )
  mode = this->DalChild<Table>::parent()->parent()->mode();

  RETURN new HiLowColumnData(
  _data[ 0 ],
  this,
  from,
  count,
  //DalSeekable<unsigned long>::from(),
  //DalSeekable<unsigned long>::count(),
  mode );

  }


  CellData *
  HiLowColumn::cellData(
  unsigned long			rowNumber,
  unsigned long			from=0,
  unsigned long			count=0,
  AccessMode			mode=AsParent ) const
  {
  PROC

  // ***********************************
  // Check that the row number is valid.
  // ***********************************
  const unsigned long rows = DalChild<Table>::parent() -> rows();

  if( rowNumber >= rows )
  {
  errstr << "The specified row number " << rowNumber << " is invalid for column: " << name() << ", which has " << rows << " rows. The row number must be in the range 0 to " << rows - 1 << "." << Sas::error( "invalidRowNumber " );
  } // if


  // ******************************************
  // Check that the column is not fixed length.
  // ******************************************
  if( cellType() == Fixed )
  {
  errstr << "The cellData() function may not be called on the fixed length column: " << name() << Sas::error( "invalidColumnOperation" );

  } // if

  if ( mode == AsParent )
  mode = this->DalChild<Table>::parent()->parent()->mode();

  RETURN new HiLowCellData(
  _data[ rowNumber ],
  this,
  rowNumber,
  from,
  count,
  mode );

  }
*/


void
HiLowColumn::insertRows( unsigned long pos, unsigned long count )
{
    PROC
    
	checkIfReadOnly();
  
    DalSeekable<unsigned long>::seek(
	this -> DalChild<Table>::parent() -> from(),
	this -> DalChild<Table>::parent() -> count() );
  
    if(cellType() == Fixed)
	_data[ 0 ]->insertBytes( pos * elements(), count * elements() );
    else {
	HiLowDataStorage ** newdata = new (HiLowDataStorage *)[rows()];
	for(unsigned long i = 0; i < rows(); ++i){
	    bool hasdata = true;
	    unsigned long k = 0;
	    if(i < pos)
		k = i;
	    else if (i >= pos + count)
		k = i - (pos + count);
	    else
		hasdata = false;

	    unsigned long size = 0;
	    CellData * s = 0;
	    if(hasdata){
		s = cellData(k);
		size = s->size();
		delete s;
		delete _data[k];
	    }
	    newdata[i] = new CellDataStorage(name(),
					     dataType(),
					     size,
					     0);
	}
	_data = newdata;
    }
    ENDPROC
	}


void
HiLowColumn::deleteRows( unsigned long from, unsigned long count )
{
    PROC
    
	checkIfReadOnly();
  
    DalSeekable<unsigned long>::seek(
	this -> DalChild<Table>::parent() -> from(),
	this -> DalChild<Table>::parent() -> count() );
  
    if(cellType() == Fixed)
	_data[ 0 ]->deleteBytes( from * elements(), count * elements() );
    else {
	HiLowDataStorage ** newdata = new (HiLowDataStorage *)[rows()];
	for(unsigned long i = 0; i < rows() + count ; ++i){
	    delete _data[i];      
	}
	for(unsigned long i = 0; i < rows(); ++i)
	    newdata[i] = new CellDataStorage(name(),
					     dataType(),
					     0,
					     0);
	_data = newdata;
    }
    ENDPROC
	}


void
HiLowColumn::copyRows( unsigned long from, unsigned long to, unsigned long count )
{
    PROC
    
	if(from != to){
	    checkIfReadOnly();
  
	    if(cellType() == Fixed){
		// **************************************************
		// This is probably not the purest way of doing this,
		// but it should work.
		// **************************************************
		HiLowColumnData * oldcd = ((ColumnDataStorage&)*(_data[0])).parent();
    
		HiLowColumnData* cd = new HiLowColumnData(
		    _data[ 0 ],
		    this,
		    0,
		    rows(),
		    AsParent );
		_data[ 0 ] -> release( 0, rows() * elements() );
    
		addDataObject( cd );
    
		((ColumnDataStorage&)*(_data[0])).parent( cd );
    
		_data[ 0 ]->copyBytes(
		    from * elements(),
		    to * elements(),
		    count * elements() );
    
		delete cd;
    
		((ColumnDataStorage&)*(_data[0])).parent( oldcd );
    
	    }  else {

		for(unsigned long i = 0; i < count; ++i){
		    unsigned long src_row_idx = from + i;
		    unsigned long des_row_idx = to + i;
		    CellData * s = cellData(src_row_idx);
		    CellData * d = cellData(des_row_idx);

		    d->size(s->size());

		    switch(dataType()){
			case Bool:
			{
			    bool8 * sd = s->boolData();
			    bool8 * dd = d->boolData();
			    for(unsigned long k = 0; k < s->size(); ++k)
				dd[k] = sd[k];
			    break;
			}
			case Int8:
			{
			    int8 * sd = s->int8Data();
			    int8 * dd = d->int8Data();
			    for(unsigned long k = 0; k < s->size(); ++k)
				dd[k] = sd[k];
			    break;
			}
			case Int16:
			{
			    int16 * sd = s->int16Data();
			    int16 * dd = d->int16Data();
			    for(unsigned long k = 0; k < s->size(); ++k)
				dd[k] = sd[k];
			    break;
			}
			case Uint16:
			{
			    uint16 * sd = s->uint16Data();
			    uint16 * dd = d->uint16Data();
			    for(unsigned long k = 0; k < s->size(); ++k)
				dd[k] = sd[k];
			    break;
			}
			case Int32:
			{
			    int32 * sd = s->int32Data();
			    int32 * dd = d->int32Data();
			    for(unsigned long k = 0; k < s->size(); ++k)
				dd[k] = sd[k];
			    break;
			}
			case Uint32:
			{
			    uint32 * sd = s->uint32Data();
			    uint32 * dd = d->uint32Data();
			    for(unsigned long k = 0; k < s->size(); ++k)
				dd[k] = sd[k];
			    break;
			}
			case Real32:
			{
			    real32 * sd = s->real32Data();
			    real32 * dd = d->real32Data();
			    for(unsigned long k = 0; k < s->size(); ++k)
				dd[k] = sd[k];
			    break;
			}
			case Real64:
			{
			    real64 * sd = s->real64Data();
			    real64 * dd = d->real64Data();
			    for(unsigned long k = 0; k < s->size(); ++k)
				dd[k] = sd[k];
			    break;
			}
			case DString:
			{
			    char * sd = s->stringData();
			    char * dd = d->stringData();
			    for(unsigned long k = 0; k < s->size(); ++k)
				dd[k] = sd[k];
			    break;
			}
		    }
      
		    delete d;
		    delete s;
		}

		release();

	    }
	}
    ENDPROC
	}


void
HiLowColumn::printOn( ostream & os ) const
{
    PROC

	// **********************
	// Write the Column Name.
	// **********************
	os << "COLUMN: Name->" << name() << " dataType=" << int( dataType() ) << "CellType=" << (cellType() == Fixed ? "Fixed" : "Variable" ) << " Label=" << label() << " units=" << units() << " dimensions=" << dimensions() << endl;

    // ****************************
    // Write the column attributes.
    // ****************************
    DalAttributable::printOn( os );

    // **********************
    // Write the Column Data.
    // **********************

    ENDPROC

	}


void
HiLowColumn::rename( const string & newName )
{
    PROC
	if( !DalChild<Table>::parent() -> hasColumn( newName ))
	{
	    // *******************
	    // Rename this object.
	    // *******************
	    DalLabelled::rename( newName );
	}
	else
	{
	    errstr << "It is not possible to rename the column with name '" << name() << "' to '" << newName << "', because a column with name '" << newName << "' already exists in the table with name '" << DalChild<Table>::parent() -> name() << "'" << Sas::error( "invalidColumnName" );
	} // if
    ENDPROC
	
	}


unsigned long
HiLowColumn::elements() const
{
    PROC
	RETURN ::numberOfElements( size() );
}


unsigned int
HiLowColumn::dataSize() const
{
    PROC
	RETURN _data[ 0 ] -> dataSize();
}


//void
//HiLowColumn::release()
//{
//	PROC
//	while( _dataObjects.size() )
//	{
//		errstr << "Unreleased data detected for column with qualified name " << qualifiedName() << Sas::warning( "unreleasedData" );
//		delete _dataObjects[ _dataObjects.size() - 1 ];
//	} // while

//if( cellType() == Fixed )
//{
//	_data[ 0 ] -> releaseAll();
//}
//else
//{
//	for( unsigned int i = 0; i < DalChild<Table>::parent() -> rows(); ++i )
//	{
//		_data[ i ] -> releaseAll();
//	} // for
//} // if
//	ENDPROC
//}


#ifdef OBSOLETE

void
HiLowColumn::scaleData(
    double			zero,
    double			scale,
    DataType		toType )
{

    PROC

	HiLowColumn		newColumn(
	    name(),
	    toType,
	    label(),
	    units(),
	    size(),
	    parent() );


    // ************************************
    // Set the following:
    //   scaling
    //   data
    // ************************************
    newColumn.set( zero, scale, toType );
    newColumn.setData( *this );

    *this = newColumn;

    ENDPROC;

}


void
HiLowColumn::unscaleData( DataType toType )
{
    PROC

	HiLowColumn		newColumn(
	    name(),
	    toType,
	    label(),
	    units(),
	    size(),
	    parent() );


    // ************************************
    // Set the following:
    //   scaling
    //   data
    // Unset
    //   scale
    // ************************************
    newColumn.set( -zero() / scale(), 1.0 / scale(), toType );
    newColumn.setData( *this );
    newColumn.unset();

    *this = newColumn;

    ENDPROC;

}


template<class T, class S> static void
copyArray(
    T *			to,
    S *			from,
    unsigned long		noOfElementsToCopy,
    DataComponent &		scale )
{
    PROC

	for( unsigned long i = 0; i < noOfElementsToCopy; ++i )
	{
	    to[ i ] = (T)((from[ i ] - scale.zero())/ scale.scale());

	} // for

    ENDPROC;

}


template< class T> static void
copyData(
    HiLowColumn &			toThisColumn,
    const T *			addressOfDataToCopy )
{
    PROC

	switch( toThisColumn.dataType() )
	{
	    case Column::Int8 :

		copyArray(
		    toThisColumn.data() -> int8Data(), 
		    addressOfDataToCopy,
		    toThisColumn.rows(),
		    toThisColumn );

		break;

	    case Column::Int16 :

		copyArray(
		    toThisColumn.data() -> int16Data(), 
		    addressOfDataToCopy,
		    toThisColumn.rows(),
		    toThisColumn );

		break;

	    case Column::Int32 :

		copyArray(
		    toThisColumn.data() -> int32Data(), 
		    addressOfDataToCopy,
		    toThisColumn.rows(),
		    toThisColumn );

		break;

	    case Column::Real32 :

		copyArray(
		    toThisColumn.data() -> real32Data(), 
		    addressOfDataToCopy,
		    toThisColumn.rows(),
		    toThisColumn );

		break;

	    case Column::Real64 :

		copyArray(
		    toThisColumn.data() -> real64Data(), 
		    addressOfDataToCopy,
		    toThisColumn.rows(),
		    toThisColumn );

		break;

	    default :

		errstr << "Encountered an invalid data type " << toThisColumn.dataType() << " in column " << toThisColumn.name() << " in table " << toThisColumn.parent()->name() << " in data set " << toThisColumn.parent()->parent()->name() << " when trying to copy column data" << Sas::error("invalidColumnDataType");

		break;

	} // switch

    ENDPROC;

}


void
HiLowColumn::setData( const HiLowColumn &	usingDataInThisColumn )
{
    PROC

	switch( usingDataInThisColumn.dataType () )
	{
	    case Column::Int8 :
		copyData(
		    *this,
		    usingDataInThisColumn.data() -> int8Data() );
		break;

	    case Column::Int16 :
		copyData(
		    *this,
		    usingDataInThisColumn.data() -> int16Data() );
		break;

	    case Column::Int32 :
		copyData(
		    *this,
		    usingDataInThisColumn.data() -> int32Data() );
		break;

	    case Column::Real32 :
		copyData(
		    *this,
		    usingDataInThisColumn.data() -> real32Data() );
		break;

	    case Column::Real64 :
		copyData(
		    *this,
		    usingDataInThisColumn.data() -> real64Data() );
		break;

	    default :

		errstr << "Encountered an invalid data type " << usingDataInThisColumn.dataType() << " in column " << usingDataInThisColumn.name() << Sas::error("invalidDataType");

		break;

	} // switch

    ENDPROC;

}

#endif

