using namespace std;
/*************************************************************** l
* 
* HiLowDataStorage.cc             & XMM Science Analysis System (SAS) 
* 
* Class representing data storage object.
* 
* (c) 1997 ESA - Written by Mark Thomas, CARA
* 
* $Id: HiLowDataStorage.cc,v 1.1.1.1 2002/08/30 13:06:29 gvacanti Exp $ 
* 
*****************************************************************/ 


#include "HiLowDataStorage.h"

#include <errstr.h>
#include <Tracer.h>

#include <string>
#include <vector>

#include "Dal.h"
#include "DalChild.h"
#include "DalSeekable.h"
#include "Slicer.h"
#include "DalDal.h"
#include "dal_utilities.h"


class HiLowValues : public DalDal, public DalChild<HiLowDataStorage>, public DalSeekable<unsigned long>
{
	public :

		// Constructors
		HiLowValues( 
			unsigned long			size,
			HiLowDataStorage *		parent );

		virtual ~HiLowValues();

		int8 * int8Values( unsigned long from, unsigned long count );
		uint16 * uint16Values( unsigned long from, unsigned long  count );
		int16 * int16Values( unsigned long from, unsigned long count );
		uint32 * uint32Values( unsigned long from, unsigned long count  );
		int32 * int32Values( unsigned long from, unsigned long count );
		real32 * real32Values( unsigned long from, unsigned long count  );
		real64 * real64Values( unsigned long from, unsigned long count  );
		bool8 * boolValues( unsigned long from, unsigned long count );
		char * stringValues( unsigned long from, unsigned long count );
		virtual unsigned int dataSize() const = 0;
		unsigned long size() const;
		void size(unsigned long size);
		virtual DataType dataType() const = 0;
		virtual void printOn( ostream& os, const string & sep="" ) const = 0;
		void insertBytes( unsigned long pos, unsigned long count );
		void deleteBytes( unsigned long from, unsigned long count );
		void copyBytes( unsigned long from, unsigned long to, unsigned long count );
		void release( unsigned long from, unsigned long count );
		void release();
		//void initialise( unsigned long from, unsigned long count );
		void initialise(
				const DalSeekable<unsigned long>& oldSeek,
				const DalSeekable<unsigned long>& seek );

	protected :
		Slicer*			_slicer;
	private :
		unsigned long		_size;

		void invalidDataType( Dal::DataType requestedType )
		{
			PROC
			errstr << "It is not possible to return the requested data (requested type = " << int( requestedType ) << "), as it is of the wrong type (actual type = " << int( dataType() ) << ")." << Sas::error( "invalidDataType" );
			ENDPROC
		}

};


template<class T>
static void writeHiLowValues(
		ostream&	os,
		T *		values,
		unsigned long	numberOfValues,
		const string &	sep="" )
{
	PROC

	for( unsigned long j = 0; j < numberOfValues; ++j )
	{
		os << values[ j ] << sep;
	} // for

	os << endl;

	ENDPROC
}


class BoolHiLowValues : public virtual HiLowValues
{
	public :

		// Constructors
		BoolHiLowValues(
			unsigned long			size,
			HiLowDataStorage *		parent );

		~BoolHiLowValues();
		unsigned int dataSize() const;
		DataType dataType() const;
		void printOn( ostream& os, const string & sep="" ) const;

};



class Int8HiLowValues : public virtual HiLowValues
{
	public :

		// Constructors
		Int8HiLowValues(
			unsigned long			size,
			HiLowDataStorage *		parent );

		~Int8HiLowValues();

		unsigned int dataSize() const;
		DataType dataType() const;
		void printOn( ostream& os, const string & sep="" ) const;


};


class Uint16HiLowValues : public virtual HiLowValues
{
	public :

		// Constructors
		Uint16HiLowValues(
			unsigned long			size,
			HiLowDataStorage *		parent );

		~Uint16HiLowValues();

		unsigned int dataSize() const;
		DataType dataType() const;
		void printOn( ostream& os, const string & sep="" ) const;


};


class Int16HiLowValues : public virtual HiLowValues
{
	public :

		// Constructors
		Int16HiLowValues(
			unsigned long			size,
			HiLowDataStorage *		parent );

		~Int16HiLowValues();

		unsigned int dataSize() const;
		DataType dataType() const;
		void printOn( ostream& os, const string & sep="" ) const;


};


class Uint32HiLowValues : public virtual HiLowValues
{
	public :

		// Constructors
		Uint32HiLowValues(
			unsigned long			size,
			HiLowDataStorage *		parent );

		~Uint32HiLowValues();

		unsigned int dataSize() const;
		DataType dataType() const;
		void printOn( ostream& os, const string & sep="" ) const;


};


class Int32HiLowValues : public virtual HiLowValues
{
	public :

		// Constructors
		Int32HiLowValues(
			unsigned long			size,
			HiLowDataStorage *		parent );

		~Int32HiLowValues();

		unsigned int dataSize() const;
		DataType dataType() const;
		void printOn( ostream& os, const string & sep="" ) const;

};


class Real32HiLowValues : public virtual HiLowValues
{
	public :

		// Constructors
		Real32HiLowValues(
			unsigned long			size,
			HiLowDataStorage *		parent );

		~Real32HiLowValues();

		unsigned int dataSize() const;
		DataType dataType() const;
		void printOn( ostream& os, const string & sep="" ) const;


};


class Real64HiLowValues : public virtual HiLowValues
{
	public :

		// Constructors
		Real64HiLowValues(
			unsigned long			size,
			HiLowDataStorage *		parent );

		~Real64HiLowValues();

		unsigned int dataSize() const;
		DataType dataType() const;
		void printOn( ostream& os, const string & sep="" ) const;


};


class StringHiLowValues : public virtual HiLowValues
{
	public :

		// Constructors
		StringHiLowValues(
			unsigned long			size,
			HiLowDataStorage *		parent );

		~StringHiLowValues();
		unsigned int dataSize() const;
		DataType dataType() const;
		void printOn( ostream& os, const string & sep="" ) const;

};


unsigned long
HiLowValues::size() const
{
	PROC
	RETURN _size;
}


void
HiLowValues::size(unsigned long size)
{
	PROC
	_size = size;
	ENDPROC
}


// *********************************
// HiLowValues methods defined here.
// *********************************

HiLowValues::HiLowValues(
	unsigned long			size,
	HiLowDataStorage *		parent )
	:
		DalDal( parent -> mode(), parent ),
		DalChild<HiLowDataStorage>( parent ),
		DalSeekable<unsigned long>( 0, 0 ),
		_slicer( new Slicer( parent )),
		_size( size )
{
	PROC
	ENDPROC
}


HiLowValues::~HiLowValues()
{
	PROC
	// ****************************
	// Release and delete the data.
	// ****************************
	release();

	delete _slicer;
	ENDPROC
}


int8 *
HiLowValues::int8Values( unsigned long from, unsigned long count )
{
	PROC
	RETURN _slicer -> slice( from, count ) -> int8Values();
}


uint16 *
HiLowValues::uint16Values( unsigned long from, unsigned long count )
{
	PROC
	RETURN _slicer -> slice( from, count ) -> uint16Values();
}


int16 *
HiLowValues::int16Values( unsigned long from, unsigned long count )
{
	PROC
	RETURN _slicer -> slice( from, count ) -> int16Values();
}


uint32 *
HiLowValues::uint32Values( unsigned long from, unsigned long  count )
{
	PROC
	RETURN _slicer -> slice( from, count ) -> uint32Values();
}


int32 *
HiLowValues::int32Values( unsigned long from, unsigned long count )
{
	PROC
	RETURN _slicer -> slice( from, count ) -> int32Values();
}


real32 *
HiLowValues::real32Values( unsigned long from, unsigned long  count )
{
	PROC
	RETURN _slicer -> slice( from, count ) -> real32Values();
}


real64 *
HiLowValues::real64Values( unsigned long from, unsigned long  count )
{
	PROC
	RETURN _slicer -> slice( from, count ) -> real64Values();
}


bool8 *
HiLowValues::boolValues( unsigned long from, unsigned long count )
{
	PROC
	RETURN _slicer -> slice( from, count ) -> boolValues();
}


char *
HiLowValues::stringValues( unsigned long from, unsigned long count )
{
	PROC
	RETURN _slicer -> slice( from, count ) -> stringValues();
}


void
HiLowValues::release( unsigned long from, unsigned long count )
{
	PROC
	_slicer -> release( from, count );
	ENDPROC
}


void
HiLowValues::release()
{
	PROC
	_slicer -> releaseAll();
	ENDPROC
}


void
HiLowValues::initialise(
	const DalSeekable<unsigned long>& oldSeek,
	const DalSeekable<unsigned long>& seek )
//void
//HiLowValues::initialise( unsigned long from, unsigned long count )
{
	PROC
	_slicer -> initialise( oldSeek, seek );
	ENDPROC
}


void
HiLowValues::insertBytes( unsigned long pos, unsigned long count )
{
	PROC

	_slicer -> insertBytes( pos, count );
	_size += count;

	ENDPROC
}


void
HiLowValues::deleteBytes( unsigned long from, unsigned long count )
{
	PROC

	_slicer -> deleteBytes( from, count );
	_size -= count;

	ENDPROC
}


void
HiLowValues::copyBytes( unsigned long from, unsigned long to, unsigned long count )
{
	PROC

	_slicer -> copyBytes( from, to, count );

	ENDPROC
}


// *************************************
// BoolHiLowValues methods defined here.
// *************************************

BoolHiLowValues::BoolHiLowValues(
	unsigned long			size,
	HiLowDataStorage *		parent )
	:
		HiLowValues( size, parent )
{
	PROC
	ENDPROC
}


BoolHiLowValues::~BoolHiLowValues()
{
	PROC
	ENDPROC
}


unsigned int
BoolHiLowValues::dataSize() const
{
	PROC
	RETURN sizeof( bool8 );
}


HiLowValues::DataType
BoolHiLowValues::dataType() const
{
	PROC
	RETURN Bool;
}


void
BoolHiLowValues::printOn( ostream& os, const string & sep) const
{
	PROC

	writeHiLowValues(
			os,
			_slicer -> slice( from(), count() ) -> boolValues(),
			//boolValues( from(), count() ),
			HiLowValues::size(),
			sep );
	ENDPROC
}


// ********************************
// Int8HiLowValues methods defined here.
// ********************************

Int8HiLowValues::Int8HiLowValues(
	unsigned long			size,
	HiLowDataStorage *		parent )
	:	HiLowValues( size, parent )
{
	PROC
	ENDPROC
}


Int8HiLowValues::~Int8HiLowValues()
{
	PROC
	ENDPROC
}


unsigned int
Int8HiLowValues::dataSize() const
{
	PROC
	RETURN sizeof( int8 );
}


HiLowValues::DataType
Int8HiLowValues::dataType() const
{
	PROC
	RETURN Int8;

}


void
Int8HiLowValues::printOn( ostream& os, const string & sep) const
{
	PROC

	writeHiLowValues(
			os,
			_slicer -> slice( from(), count() ) -> int8Values(),
			//int8Values( from, count() ),
			HiLowValues::size(),
			sep );

	ENDPROC
}


// ***************************************
// Uint16HiLowValues methods defined here.
// ***************************************

Uint16HiLowValues::Uint16HiLowValues(
	unsigned long			size,
	HiLowDataStorage *		parent )
	:	HiLowValues( size, parent )
{
	PROC
	ENDPROC
}


Uint16HiLowValues::~Uint16HiLowValues()
{
	PROC
	ENDPROC
}


unsigned int
Uint16HiLowValues::dataSize() const
{
	PROC
	RETURN sizeof( uint16 );
}


HiLowValues::DataType
Uint16HiLowValues::dataType() const
{
	PROC
	RETURN Uint16;
}


void
Uint16HiLowValues::printOn( ostream& os, const string & sep) const
{
	PROC

	writeHiLowValues(
			os,
			_slicer -> slice( from(), count() ) -> uint16Values(),
			HiLowValues::size(),
			sep );

	ENDPROC

}


// *********************************
// Int16HiLowValues methods defined here.
// *********************************

Int16HiLowValues::Int16HiLowValues(
	unsigned long			size,
	HiLowDataStorage *		parent )
	:	HiLowValues( size, parent )
{
	PROC
	ENDPROC
}


Int16HiLowValues::~Int16HiLowValues()
{
	PROC
	ENDPROC
}


unsigned int
Int16HiLowValues::dataSize() const
{
	PROC
	RETURN sizeof( int16 );
}


HiLowValues::DataType
Int16HiLowValues::dataType() const
{
	PROC
	RETURN Int16;
}


void
Int16HiLowValues::printOn( ostream& os, const string & sep) const
{
	PROC

	writeHiLowValues(
			os,
			_slicer -> slice( from(), count() ) -> int16Values(),
			HiLowValues::size(),
			sep );

	ENDPROC

}


// ***************************************
// Uint32HiLowValues methods defined here.
// ***************************************

Uint32HiLowValues::Uint32HiLowValues(
	unsigned long			size,
	HiLowDataStorage *		parent )
	:	HiLowValues( size, parent )
{
	PROC
	ENDPROC
}


Uint32HiLowValues::~Uint32HiLowValues()
{
	PROC
	ENDPROC
}


unsigned int
Uint32HiLowValues::dataSize() const
{
	PROC
	RETURN sizeof( uint32 );
}


HiLowValues::DataType
Uint32HiLowValues::dataType() const
{
	PROC
	RETURN Uint32;
}


void
Uint32HiLowValues::printOn( ostream& os, const string & sep) const
{
	PROC

	writeHiLowValues(
			os,
			_slicer -> slice( from(), count() ) -> uint32Values(),
			HiLowValues::size(),
			sep );

	ENDPROC
}



// *********************************
// Int32HiLowValues methods defined here.
// *********************************

Int32HiLowValues::Int32HiLowValues(
	unsigned long			size,
	HiLowDataStorage *		parent )
	:	HiLowValues( size, parent )
{
	PROC
	ENDPROC
}


Int32HiLowValues::~Int32HiLowValues()
{
	PROC
	ENDPROC
}


unsigned int
Int32HiLowValues::dataSize() const
{
	PROC
	RETURN sizeof( int32 );
}


HiLowValues::DataType
Int32HiLowValues::dataType() const
{
	PROC
	RETURN Int32;
}


void
Int32HiLowValues::printOn( ostream& os, const string & sep) const
{
	PROC

	writeHiLowValues(
			os,
			_slicer -> slice( from(), count() ) -> int32Values(),
			HiLowValues::size(),
			sep );

	ENDPROC
}



// **********************************
// Real32HiLowValues methods defined here.
// **********************************

Real32HiLowValues::Real32HiLowValues(
	unsigned long			size,
	HiLowDataStorage *		parent )
	:	HiLowValues( size, parent )
{
	PROC
	ENDPROC
}


Real32HiLowValues::~Real32HiLowValues()
{
	PROC
	ENDPROC
}


unsigned int
Real32HiLowValues::dataSize() const
{
	PROC
	RETURN sizeof( real32 );
}


HiLowValues::DataType
Real32HiLowValues::dataType() const
{
	PROC
	RETURN Real32;
}


void
Real32HiLowValues::printOn( ostream& os, const string & sep) const
{
	PROC

	writeHiLowValues(
			os,
			_slicer -> slice( from(), count() ) -> real32Values(),
			HiLowValues::size(),
			sep );

	ENDPROC
}



// **********************************
// Real64HiLowValues methods defined here.
// **********************************

Real64HiLowValues::Real64HiLowValues(
	unsigned long			size,
	HiLowDataStorage *		parent )
	:	HiLowValues( size, parent )
{
	PROC
	ENDPROC
}


Real64HiLowValues::~Real64HiLowValues()
{
	PROC
	ENDPROC
}


unsigned int
Real64HiLowValues::dataSize() const
{
	PROC
	RETURN sizeof( real64 );
}


HiLowValues::DataType
Real64HiLowValues::dataType() const
{
	PROC
	RETURN Real64;
}


void
Real64HiLowValues::printOn( ostream& os, const string & sep) const
{
	PROC

	writeHiLowValues(
			os,
			_slicer -> slice( from(), count() ) -> real64Values(),
			HiLowValues::size(),
			sep );

	ENDPROC
}



// ***************************************
// StringHiLowValues methods defined here.
// ***************************************

StringHiLowValues::StringHiLowValues(
	unsigned long			size,
	HiLowDataStorage *		parent )
	:	HiLowValues( size, parent )
{
	PROC
	ENDPROC
}


StringHiLowValues::~StringHiLowValues()
{
	PROC
	ENDPROC
}


unsigned int
StringHiLowValues::dataSize() const
{
	PROC
	RETURN sizeof( dstring );
}


HiLowValues::DataType
StringHiLowValues::dataType() const
{
	PROC
	RETURN DString;
}


void
StringHiLowValues::printOn( ostream& os, const string & sep) const
{
	PROC

	writeHiLowValues(
			os,
			_slicer -> slice( from(), count() ) -> stringValues(),
			HiLowValues::size(),
			sep );

	ENDPROC
}


// *********************************
// HiLowDataStorage methods defined here.
// *********************************

HiLowDataStorage::HiLowDataStorage(
			const string&			name,
			DataType			dataType,
			unsigned long			size )
	:
		DalDal( Dal::Read, 0 ),
		_name(name)
{
	PROC

	SAS_DEBUG( clog << "Constructing HiLowDataStorage name=" << name << " type=" << int( dataType ) << " size=" << size << endl; )

	switch( dataType )
	{
		case Bool :
			_values = new BoolHiLowValues( size, this );
			break;
		case Int8 :
			_values = new Int8HiLowValues( size, this );
			break;
		case Uint16 :
			_values = new Uint16HiLowValues( size, this );
			break;
		case Int16 :
			_values = new Int16HiLowValues( size, this );
			break;
		case Uint32 :
			_values = new Uint32HiLowValues( size, this );
			break;
		case Int32 :
			_values = new Int32HiLowValues( size, this );
			break;
		case Real32 :
			_values = new Real32HiLowValues( size, this );
			break;
		case Real64 :
			_values = new Real64HiLowValues( size, this );
			break;
		case DString :
			_values = new StringHiLowValues( size, this );
			break;
		default :
			errstr << "encountered invalid type " << int( dataType)<< " for data storage of " << _name << Sas::error("internalError");
			break;
	} // switch

	ENDPROC
}



HiLowDataStorage::~HiLowDataStorage()
{
	PROC

	// ****************************
	// Release and delete the data.
	// ****************************
	//release();

	//delete _values;

	ENDPROC
}

void HiLowDataStorage::cannotConvertTo(const string& type) const
{
	errstr << "Accessing data " << _name << " of type " << toString( _values -> dataType() ) << " as " << type << " is not supported" << Sas::error("accessingAsWrongDataType");
}


int8 *
HiLowDataStorage::int8Data( unsigned long from, unsigned long count )
{
	PROC

	if( _values -> dataType() != Int8 )
	{
		// The following code used to convert the data
		// to the type asked. This is abandoned, since
		// the old data type is irrevocably lost and
		// the new data type will be written to the output.
		//
		//HiLowValues * values = new Int8HiLowValues(
		//			_values -> rows(),
		//			_values -> size() );
		//
		//_values -> getHiLowValues( values -> int8HiLowValues() );
		//delete _values;
		//_values = values;

		//BUG: should report name of object.
		//errstr << "Accessing data " << _name << " of type " << int( _values -> dataType() ) << " as int8 is not supported" << Sas::error("accessingAsWrongDataType");
		cannotConvertTo("int8");

	} // if

	RETURN _values -> int8Values( from, count );

}


uint16 *
HiLowDataStorage::uint16Data( unsigned long from, unsigned long count )
{
	PROC

	if( _values -> dataType() != Uint16 )
	{
		cannotConvertTo("uint16");
	} // if

	RETURN _values -> uint16Values( from, count );

}


int16 *
HiLowDataStorage::int16Data( unsigned long from, unsigned long count )
{
	PROC

	if( _values -> dataType() != Int16 )
	{
		cannotConvertTo("int16");
	} // if

	RETURN _values -> int16Values( from, count );

}


uint32 *
HiLowDataStorage::uint32Data( unsigned long from, unsigned long count )
{
	PROC

	if( _values -> dataType() != Uint32 )
	{
		cannotConvertTo("uint32");
	} // if

	RETURN _values -> uint32Values( from, count );

}


int32 *
HiLowDataStorage::int32Data( unsigned long from, unsigned long count )
{
	PROC

	if( _values -> dataType() != Int32 )
	{
		cannotConvertTo("int32");
	} // if

	RETURN _values -> int32Values( from, count );

}


real32 *
HiLowDataStorage::real32Data( unsigned long from, unsigned long count )
{
	PROC

	if( _values -> dataType() != Real32 )
	{
		cannotConvertTo("real32");
	} // if

	RETURN _values -> real32Values( from, count );

}


real64 *
HiLowDataStorage::real64Data( unsigned long from, unsigned long count )
{
	PROC

	if( _values -> dataType() != Real64 )
	{
		cannotConvertTo("real64");
	} // if

	RETURN _values -> real64Values( from, count );

}


bool8 *
HiLowDataStorage::boolData( unsigned long from, unsigned long count )
{
	PROC

	if( _values -> dataType() != Bool )
	{
		// *********************************
		// Convert Data to appropriate type.
		// *********************************

		// ****************
		// NOT IMPLEMENTED.
		// ****************
		//BUG: should report name of object.
		cannotConvertTo("string");
	} // if

	RETURN _values -> boolValues( from, count );

}


char *
HiLowDataStorage::stringData( unsigned long from, unsigned long count )
{
	PROC

	if( _values -> dataType() != DString )
	{
		cannotConvertTo("string");
	} // if

	RETURN _values -> stringValues( from, count );

}


unsigned int
HiLowDataStorage::dataSize() const
{
	PROC RETURN _values -> dataSize();

}


HiLowDataStorage::DataType
HiLowDataStorage::dataType() const
{
	PROC
	RETURN _values -> dataType();

}


unsigned long
HiLowDataStorage::size() const
{
	PROC
	RETURN _values -> size();
}


void
HiLowDataStorage::size(unsigned long size )
{
	PROC
	_values -> size( size );
	ENDPROC
}


unsigned long
HiLowDataStorage::from() const
{
	PROC
	RETURN _values -> from();
}


unsigned long
HiLowDataStorage::count() const
{
	PROC
	RETURN _values -> count();
}


void
HiLowDataStorage::insertBytes( unsigned long pos, unsigned long count )
{
	PROC

	_values -> insertBytes( pos, count );

	ENDPROC
}


void
HiLowDataStorage::deleteBytes( unsigned long from, unsigned long count )
{
	PROC

	_values -> deleteBytes( from, count );

	ENDPROC
}


void
HiLowDataStorage::copyBytes( unsigned long from, unsigned long to, unsigned long count )
{
	PROC

	_values -> copyBytes( from, to, count );

	ENDPROC
}


void
HiLowDataStorage::printOn( ostream & os, const string & sep)
{
	PROC

	_values -> printOn( os, sep );

	ENDPROC
}


void
HiLowDataStorage::release( unsigned long from, unsigned long count )
{
	PROC
	_values -> release( from, count );
	ENDPROC
}


void
HiLowDataStorage::release()
{
	PROC
	_values -> release();
	ENDPROC
}


void
HiLowDataStorage::initialise(
	const DalSeekable<unsigned long>& oldSeek,
	const DalSeekable<unsigned long>& seek )
//void
//HiLowDataStorage::initialise( unsigned long from, unsigned long count )
{
	PROC
	_values -> initialise( oldSeek, seek );
	ENDPROC
}


