/*************************************************************** 
* 
* FitsReaderWriter.h     & XMM Science Analysis System (SAS) 
* 
* Class representing FITS file reader/writer.
* 
* (c) 1997 ESA - Written by Mark Thomas, CARA
* 
* $Id: FitsReaderWriter.h,v 1.1.1.1 2002/08/30 13:06:29 gvacanti Exp $ 
* 
*****************************************************************/ 
#ifndef FITS_READER_WRITER_H
#define FITS_READER_WRITER_H


#include <fitsio.h>
#include <map>
#include <stringutil.h>


#include "dal_utilities.h"
#include "DataSetReaderWriter.h"


class FitsReaderWriter : public virtual DataSetReaderWriter
{
	public :
		virtual ~FitsReaderWriter(){}
		//DataSetServer::FileType fileType() const { PROC RETURN DataSetServer::Fits; }

		//static std::string magic() { PROC RETURN "SIMPLE"; }
		static bool exists( const std::string& setName );

	private :
		static std::string magic() { PROC RETURN "SIMPLE"; }
		static const std::string attributeSeparatorComment;
		static const std::string defaultColumnComment;
		static const std::string arrayUnitsComment;
		static const std::string columnNullValueComment;
		static const std::string arrayNullValueComment;

	protected :

		struct AttributeInfo
		{
			std::string					_name;
			std::string					_label;
			std::string					_units;
			Attribute::DataType			_dataType;
			std::string					_value;
		};
		struct ColumnInfo
		{
			ColumnInfo() :
				_nullValueFound( false ),
				_scalingFactorsFound( false ) {}
			fitsfile*				_fptr;
			unsigned int				_columnNumber;
			std::string					_name;
			std::string					_label;
			std::string					_units;
			Column::DataType			_dataType;
			int					_typeCode;
			long int				_repeat;
			long int				_width;
			longVec					_size;
			unsigned int				_numberOfAttributes;
			std::vector<AttributeInfo>			_attributes;
			bool					_nullValueFound;
			int					_nullValue;
			bool					_scalingFactorsFound;
			double					_scale;
			double					_zero;
		};
		struct TableInfo
		{
			fitsfile*				_fptr;
			std::string					_name;
			std::string					_label;
			unsigned long				_rows;
			unsigned int				_numberOfAttributes;
			std::vector<AttributeInfo>			_attributes;
			unsigned int				_numberOfColumns;
			std::map<unsigned int, ColumnInfo>		_columns;
			std::vector<std::string>				_comments;
			std::vector<std::string>				_history;
		};

		enum Status { Good, Bad, Ignore };
		virtual Status columnDataTypeStatus( int typeCode ) const;
		virtual void setColumnDataType( ColumnInfo& columnInfo ) const;

		int readTableInfo(
			fitsfile*				fptr,
			//const DataSet&				dataSet,
			TableInfo& 				tableInfo,
			int&					status ) const;

		virtual Table* addTable(
			DataSet&				dataSet,
			TableInfo& 				tableInfo ) const;

		static const std::string attributeSeparator;

		class KeywordChecker
		{
			public :
				virtual bool ok( const std::string & name ) { PROC RETURN name == name; }
				virtual std::string trim( const std::string & name ) const { PROC RETURN name; }
				virtual void reset() { PROC ENDPROC }
		};

		class ColumnKeywordChecker : public virtual KeywordChecker
		{
			private :
				const Column * _column;
			public :
				ColumnKeywordChecker( const Column * column ) : _column( column ) { PROC ENDPROC }
				bool ok( const std::string & name )
				{
 					PROC
					bool	result = false;
					// *****************************************************************
					// Column Attributes must be one of the allowable column attributes.
					// *****************************************************************
					unsigned int columnNr = _column -> parent() -> columnNumber( _column -> name() );

					if(
						(trimDigits( name ) + toString( columnNr + 1 ) == name )
						&&
						isFITSColumnKeyword( trimDigits( name )))
					{
						result = true;
					} // if
					RETURN result;

				}
				std::string trim( const std::string & name ) const { PROC RETURN trimDigits( name ); }
		};

		class TableKeywordChecker : public virtual KeywordChecker
		{
			public :
				TableKeywordChecker( const Table * table ) : _table( table ) { PROC ENDPROC }
				bool ok( const std::string & name )
				{
					PROC
					bool	result = true;

					// ***********************************************
					// Table Attributes must not be a reserved
					// column attribute.
					// ***********************************************
					if( isFITSColumnKeyword( trimDigits( name )))
					{
						result = false;
					} // if
					RETURN result;
				}
				//std::string trim( const std::string & name ) const { PROC RETURN name; }
			private :
				const Table * _table;
		};


		class PrimaryKeywordChecker : public virtual KeywordChecker
		{
			public :
				PrimaryKeywordChecker() : _ok( false ) { PROC ENDPROC }
				void reset() { PROC _ok = false; ENDPROC }
				bool ok( const std::string & name )
				{
					PROC
					if( _ok == false && name == attributeSeparator )
					{
						_ok = true;
						ALTRETURNV( false );
					} // if
					RETURN _ok;
				}
			private :
				bool _ok;
		};


		class DataSetKeywordChecker : public virtual KeywordChecker
		{
			public :
				DataSetKeywordChecker() : _ok( true ) { PROC ENDPROC }
				void reset() { PROC _ok = true; ENDPROC }
				bool ok( const std::string & name )
				{
					PROC

					if( _ok == true && name == attributeSeparator )
					{
						_ok = false;
					} // if
					RETURN _ok;
				}
			private :
				bool _ok;
		};

		int writeAttributes( 
			fitsfile *		fptr,
			const Attributable &	attributeSet,
			const std::string &		appendThisString,
			//bool (*attributeOK)(const std::string &),
			bool (FitsReaderWriter::*attributeOK)(const std::string &) const,
			int &			status ) const;

		int writeComments( 
			fitsfile *		fptr,
			const Attributable &	attributeSet,
			int &			status ) const;

		int writeHistory( 
			fitsfile *		fptr,
			const Attributable &	attributeSet,
			int &			status ) const;

		int writeHeader(
			fitsfile *		fptr,
			const Attributable &	attributeSet,
			const std::string &		appendThisString,
			bool (FitsReaderWriter::*attributeOK)(const std::string &) const,
			//bool (*attributeOK)(const std::string &),
			int &			status ) const;

		int readCreatorAndDate(
			fitsfile*		fptr,
			const DataSet*		dataSet,
			int&			status ) const;

		int writeCreatorAndDate(
			fitsfile*		fptr,
			const DataSet*		dataSet,
			int&			status ) const;

		int readProcessHistory(
			fitsfile*		fptr,
			const DataSet*		dataSet,
			int&			status ) const;

		int readDebugHistory(
			fitsfile*		fptr,
			const DataSet*		dataSet,
			int&			status ) const;

		int writeProcessHistory(
			fitsfile*		fptr,
			const DataSet*		dataSet,
			int&			status ) const;

		int writeDebugHistory(
			fitsfile*		fptr,
			const DataSet*		dataSet,
			int&			status ) const;

		int writeSeparator(
			fitsfile *		fptr,
			unsigned int		nrOfKeywords,
			int &			status ) const;

		void setAttributeValue(
			Attribute *		attribute,
			const std::string&		value) const;

		int readAttributes( 
			fitsfile *		fptr,
			Attributable &		attributeSet,
			KeywordChecker *	attributeChecker,
			int &			status ) const;

		int readCommentsAndHistory(
			fitsfile *		fptr,
			Attributable &		attributeSet,
			KeywordChecker *	attributeChecker,
			int &			status ) const;

		int readHeader(
			fitsfile *		fptr,
			Attributable &		attributeSet,
			KeywordChecker *	attributeChecker,
			int &			status ) const;

		int writeColumnScalingFactors(
			fitsfile *	fptr,
			const Column *	column,
			int		columnNumber,
			int &		status ) const;

		int writeColumnNullValue(
			fitsfile *	fptr,
			const Column *	column,
			int &		status ) const;

		int writeColumnComment(
			fitsfile *	fptr,
			const std::string &	comment,
			unsigned int	columnNr,
			int &		status ) const;

		bool columnAttributeOK( const std::string & name ) const;

		bool tableAttributeOK( const std::string & attributeName ) const;

		int writeArrayUnits(
			fitsfile *	fptr,
			const Array &	array,
			int &		status ) const;

		int checkAndSetUnsignedArray(
			fitsfile *		fptr,
			Array::DataType&	dataType,
			int &			status ) const;

		void checkAndSetUnsignedColumn(
			ColumnInfo&		columnInfo ) const;

		//int checkAndSetUnsignedColumn(
		//	fitsfile *		fptr,
		//	int			columnNumber,
		//	Column::DataType&	dataType,
		//	int &			status ) const;

		int getArrayDataType(
			fitsfile *		fptr,
			int&			bitpix,
			Array::DataType&	dataType,
			int &			status ) const;

		int getColumnDataType(
			fitsfile *		fptr,
			int			columnNumber,
			int&			typeCode,
			//long&			repeat,
			//long&			width,
			Column::DataType&	dataType,
			int &			status ) const;
	
		int readColumnScalingFactors(
			fitsfile *	fptr,
			int		columnNumber,
			Column &	column,
			int &		status ) const;

		int readColumnNullValue(
			fitsfile *	fptr,
			Column &	column,
			int &		status ) const;

		int readArrayScalingFactors(
			fitsfile	* fptr,
			Array &		array,
			int &		status ) const;

		int readArrayNullValue(
			fitsfile *	fptr,
			Array&		array,
			int &		status ) const;

		int getBlockNameAndComment(
			fitsfile*		fptr,
			const unsigned int&	blockNumber,
			std::string&			name,
			std::string&			comment,
			int&			status ) const;

		int readColumnUnits(
			fitsfile *	fptr,
			int 		columnNumber,
			char *		units,
			char *		comment,
			int &		status ) const;

		int readArrayDimensions(
			fitsfile *	fptr,
			int &		dimensions,
			long *		size,
			int &		status ) const;

		int readArrayUnits(
			fitsfile *	fptr,
			std::string &	units,
			int &		status ) const;

		virtual bool acceptTableKeyword(
			fitsfile*	fptr,
			const std::string&	keyName ) const;

		virtual int writeColumns(
			fitsfile *	fptr,
			const Table &	table,
			int &		status ) const = 0;

		virtual int writeTable( 
			fitsfile *	fptr,
			const Table &	table,
			int &		status ) const = 0;

		//virtual int readColumn(
		//	fitsfile *	fptr,
		//	int &		columnNumber,
		//	Table &		table,
		//	int &		status ) const = 0;

		int readTable(
			fitsfile * 	fptr,
			DataSet &	dataSet,
			TableInfo& 	tableInfo,
			int &		status ) const;

		virtual int readArray(
			fitsfile *	fptr,
			DataSet &	dataSet,
			int &		status ) const = 0;

		virtual int readPrimaryArray(
			fitsfile *		fptr,
			DataSet &		dataSet,
			KeywordChecker *	keywordChecker,
			int &			status ) const = 0;

		virtual int writeArray(
			fitsfile *	fptr,
			const Array &	array,
			int &		status ) const = 0;

		int writeArrayNullValue(
			fitsfile *	fptr,
			const Array *	column,
			int &		status ) const;

		int writeLongStringWarning(
			fitsfile *	fptr,
			int &		status ) const;

		int readLongString(
			fitsfile *	fptr,
			int&		keyNumber,
			char*		value,
			char*		comment,
			int &		status ) const;

};



#endif
