using namespace std;
/*************************************************************** 
* 
* HiLowDalReaderwriter.cc     & XMM Science Analysis System (SAS) 
* 
* Class representing DAL Dataset reader/writer.
* 
* (c) 1997 ESA - Written by Mark Thomas, CARA
* 
* $Id: HiLowDalReaderWriter.cc,v 1.2 2002/09/06 14:57:38 gvacanti Exp $
* 
*****************************************************************/ 


#include <string>
#include <fstream>

#include <errstr.h>
#include <Tracer.h>

#include "Dal.h"
#include "HiLowDalReaderWriter.h"
#include "utilities.h"


static void
writeComments(
	fstream	&			fileStream,
	const Attributable &		attributeSet )
{

	PROC

	// ***************
	// write Comments.
	// ***************

	// =========================
	// write number of comments.
	// =========================
	const unsigned int numberOfComments = attributeSet.comments();

	fileStream.write(
			(char*)&numberOfComments,
			sizeof( numberOfComments ));


	// ==========================
	// write each comment in turn.
	// ==========================
	for(
		unsigned int i = 0;
		i < attributeSet.comments();
		++i )
	{
		const Attribute & attribute = *attributeSet.attribute( i );

		fileStream.write(
				(char *)attribute.name().c_str(),
				attribute.name().length() );
				
		fileStream << '\n';

	} // for

	ENDPROC

}


static void
writeHistory(
	fstream	&			fileStream,
	const Attributable &		attributeSet )
{

	PROC

	// ***************
	// write Comments.
	// ***************

	// ==============================
	// write number of history lines.
	// ==============================
	const unsigned int numberOfHistory = attributeSet.historys();

	fileStream.write(
			(char*)&numberOfHistory,
			sizeof( numberOfHistory ));


	// ================================
	// write each history line in turn.
	// ================================
	for(
		unsigned int i = 0;
		i < attributeSet.historys();
		++i )
	{
		const Attribute & attribute = *attributeSet.attribute( i );

		fileStream.write(
				(char *)attribute.name().c_str(),
				attribute.name().length() );
				
		fileStream << '\n';

	} // for

	PROC

}


static void
readComments(
	fstream	&		fileStream,
	Attributable &		attributeSet )
{

	PROC

	char			comment[ 1024 ];
	unsigned int		numberOfComments;

	// **************
	// read Comments.
	// **************

	// ========================
	// read number of comments.
	// ========================
	fileStream.read(
			(char*)&numberOfComments,
			sizeof( numberOfComments ));

	// ==================================
	// read and set each comment in turn.
	// ==================================
	for( unsigned int i = 0; i < numberOfComments; i++ )
	{

		fileStream.getline( comment, 1020 );
		attributeSet.addComment( comment );

	} // for

	PROC

}


static void
readHistory(
	fstream	&			fileStream,
	Attributable &		attributeSet )
{

	PROC

	char			history[ 1024 ];
	unsigned int		numberOfHistory;

	// *************
	// read History.
	// *************

	// =============================
	// read number of history lines.
	// =============================
	fileStream.read(
			(char*)&numberOfHistory,
			sizeof( numberOfHistory ));

	// ==================================
	// read and set each history in turn.
	// ==================================
	for( unsigned int i = 0; i < numberOfHistory; i++ )
	{

		fileStream.getline( history, 1020 );
		attributeSet.addHistory( history );

	} // for

	PROC

}


static void
writeAttributes(
	fstream	&			fileStream,
	const Attributable &		attributeSet )
{

	PROC

	// *****************
	// write Attributes.
	// *****************

	// ===========================
	// write number of attributes.
	// ===========================
	const unsigned int numberOfAttributes = attributeSet.attributes();

	fileStream.write(
			(char*)&numberOfAttributes,
			sizeof( numberOfAttributes ));

	for(
		unsigned int i = 0;
		i < attributeSet.attributes();
		++i )
	{
		const Attribute & attribute = *attributeSet.attribute( i );

		// ================
		// write attribute.
		// ================

		// ---------------------
		// write attribute name.
		// ---------------------
		fileStream.write(
			attribute.name().c_str(),
			attribute.name().length() );

		fileStream << '\n';

		// ---------------------------
		// write attribute value type.
		// ---------------------------
		const Attribute::DataType type = attribute.dataType();

		fileStream.write(
			(char*)&type,
			sizeof( type ));

		// ----------------------
		// write attribute value.
		// ----------------------
		fileStream.write(
			(char*)attribute.addressOfValue(),
			attribute.dataSize() );

		// ----------------------
		// write attribute units.
		// ----------------------
		fileStream.write(
			attribute.units().c_str(),
			attribute.units().length() );

		fileStream << '\n';

		// ------------------------
		// write attribute comment.
		// ------------------------
		fileStream.write(
			attribute.label().c_str(),
			attribute.label().length() );

		fileStream << '\n';

	} // for

	ENDPROC

}


static void
readAttributes(
	fstream	&			fileStream,
	Attributable &		attributeSet )
{

	PROC

	unsigned int numberOfAttributes;


	// ****************
	// read Attributes.
	// ****************

	// ==========================
	// read number of attributes.
	// ==========================
	fileStream.read(
			(char*)&numberOfAttributes,
			sizeof( numberOfAttributes ));

	// ====================================
	// read and set each attribute in turn.
	// ====================================
	for( unsigned int i = 0; i < numberOfAttributes; i++ )
	{
		char			attributeName[ 1024 ];
		char			attributeUnits[ 1024 ];
		char			attributeComment[ 1024 ];
		Attribute::DataType	dataType;

		fileStream.getline( attributeName, 1020 );
		fileStream.read((char*)&dataType, sizeof( dataType ));
		fileStream.getline( attributeComment, 1020 );

		// =====
		// Units
		// =====
		fileStream.getline( attributeUnits, 1020 );

		// ==========================
		// Now add the new attribute.
		// ==========================
		Attribute *	attribute =

					attributeSet.addAttribute(
							attributeName,
							attributeComment,
							attributeUnits );

		attribute->dataType(dataType);

		fileStream.read(
				(char*)attribute -> addressOfValue(),
				attribute -> dataSize() );


	} // for

	ENDPROC

}


static void
writeHeader(
	fstream	&			fileStream,
	const Attributable &		attributeSet )
{

	PROC

	writeAttributes( fileStream, attributeSet );
	writeComments( fileStream, attributeSet  );
	writeHistory( fileStream, attributeSet  );

	ENDPROC;

}


static void
readHeader(
	fstream	&			fileStream,
	Attributable &		attributeSet )
{

	PROC

	readAttributes( fileStream, attributeSet );
	readComments( fileStream, attributeSet );
	readHistory( fileStream, attributeSet );

	ENDPROC;

}


static void
writeTable(
	fstream	&		fileStream,
	const Table & 		table )
{

	PROC

	string			tableName;
	unsigned int		numberOfColumns;
	unsigned int		numberOfRows;
	Dal::DataType		dataType;
	unsigned int		numberOfDimensions;


	// ********************
	// write the Table name.
	// ********************
	tableName = table.name();
	fileStream.write(
		tableName.c_str(),
		tableName.length() );
	fileStream << '\n';

	// *************
	// write Header.
	// *************
	writeHeader( fileStream, (const Attributable&)table );


	// *****************************************
	// write the number of columns in the table.
	// *****************************************
	numberOfColumns = table.columns();

	fileStream.write(
			(char *)&numberOfColumns,
			sizeof( numberOfColumns ));

	numberOfRows = table.rows();

	fileStream.write(
			(char*) &numberOfRows,
			sizeof( numberOfRows ));


	// ********************************
	// Do for each column in the table:
	//   name
	//   data type
	//   units
	//   number of dimensions
	//   dimensions
	//   data
	// ********************************
	for( unsigned int ic = 0; ic < numberOfColumns; ic++ )
	{
		const Column *	column = table.column( ic );

		// ====
		// name
		// ====
		fileStream.write(
			(char*)column -> name().c_str(),
			column -> name().length() );
		fileStream << '\n';

		// =========
		// data type
		// =========
		dataType = column -> dataType(),

		fileStream.write(
			(char*)&dataType,
			sizeof( dataType ));

		// =====
		// Units
		// =====
		fileStream.write(
			(char*)column -> units().c_str(),
			column -> units().length() );
		fileStream << '\n';

		// ==============================================
		// Number of dimensions. The number of dimensions
		// may be zero.
		// ==============================================
		numberOfDimensions = column -> dimensions();

		fileStream.write(
			(char *)&numberOfDimensions,
			sizeof( numberOfDimensions ));
				
		// ==========
		// Dimensions
		// ==========
		fileStream.write(
			(char *)Mvec<unsigned long>( column -> size() ).data(),
			numberOfDimensions *
			sizeof( unsigned long ) );

		// ====
		// Data
		// ====
		fileStream.write(
			(char*)column -> data() -> int8Data(),
			column -> elements() * column -> dataSize() );


	} /* for */

	ENDPROC;

}


static void
readTable(
	fstream	&	fileStream,
	DataSet &	dataSet )
{
	PROC

	unsigned int		numberOfColumns;
	unsigned long		numberOfRows;
	char			tableName[ 1024 ];


	// ****************************************************
	// Do:
	//	The table Name - read
	//	The nunber of rows in the table columns
	// ****************************************************
	fileStream.getline( tableName, 1020 );

	fileStream.read(
			(char*)&numberOfRows,
			sizeof( numberOfRows ));

	// *****************************************
	// Add the table to the end of the data set.
	// *****************************************
	Table * table = dataSet.addTable( tableName, numberOfRows );

	// ************
	// read Header.
	// ************
	readHeader( fileStream, (Attributable&)*table );


	// ***********************
	// read number of columns.
	// ***********************
	fileStream.read(
			(char*)&numberOfColumns,
			sizeof( numberOfColumns ));



	// ********************************
	// Do for each column in the table:
	//   name
	//   datatype
	//   units
	//   number of dimensions
	//   dimensions
	//   data
	// ********************************
	for( unsigned int ic = 0; ic < numberOfColumns; ic++ )
	{
		char			columnName[ 1024 ];
		char			comment[ 1024 ];
		Dal::DataType		dataType;
		char			units[ 1024 ];
		unsigned int		numberOfDimensions;

		// ====
		// name
		// ====
		fileStream.getline( columnName, 1020 );

		// =========
		// data type
		// =========
		fileStream.read(
				(char*)&dataType,
				sizeof( dataType ));

		// =======
		// comment
		// =======
		fileStream.getline( comment, 1020 );

		// =====
		// units
		// =====
		fileStream.getline( units, 1020 );

		// ====================
		// number of dimensions
		// ====================
		fileStream.read(
				(char*)&numberOfDimensions,
				sizeof( numberOfDimensions ));

		// ==========
		// dimensions
		// ==========
		unsigned long	dimensions[ numberOfDimensions ];

		fileStream.read(
			(char *)dimensions,
			numberOfDimensions * sizeof( unsigned long ));

		longVec		size(
						dimensions,
						dimensions + numberOfDimensions );

		// ============================
		// Add the column to the table.
		// ============================
		Column * column = table -> addColumn(
						columnName,
						dataType,
						comment,
						units,
						size );

		// =====
		// data.
		// =====
		fileStream.read(
			(char*)column -> data() -> int8Data(),
			column -> dataSize() * column -> elements() );

	} /* for */


	ENDPROC;

}


static void
writeArray(
	fstream &	fileStream,
	const Array &	array )
{
	PROC

	// *********************
	// write the Array name.
	// *********************
	fileStream.write(
		array.name().c_str(),
		array.name().length() );
	fileStream << '\n';


	// *************
	// write Header.
	// *************
	writeHeader( fileStream, (const Attributable&)array );


	// ****************************
	// write:
	//	Data type.
	//	Number of dimensions
	//	Size of each dimension
	//	Data
	// ****************************

	// =========
	// Data Type
	// =========
	Array::DataType dataType = array.dataType();
	fileStream.write( (char*)&dataType, sizeof( dataType ));

	// ====================
	// Number of dimensions
	// ====================
	int numberOfDimensions = array.dimensions();
	fileStream.write( (char*)&numberOfDimensions, sizeof( numberOfDimensions ));

	// ==========
	// Dimensions
	// ==========
	fileStream.write(
		(char *)Mvec<unsigned long>( array.size() ).data(),
		numberOfDimensions *
		sizeof( unsigned long ) );

	// ====
	// Data
	// ====
	fileStream.write(
			(char*)array.data() -> int8Data(),
			array.elements() * array.dataSize() );

	ENDPROC;

}


static void
readArray(
	fstream &	fileStream,
	DataSet &	dataSet )
{

	PROC


	// ********************
	// read the array name.
	// ********************
	char			arrayName[ 1024 ];

	fileStream.getline( arrayName, 1020 );


	// *********
	// data type
	// *********
	Array::DataType		dataType;

	fileStream.read(
			(char*)&dataType,
			sizeof( dataType ));

	// ********************
	// Number of dimensions
	// ********************
	unsigned int		numberOfDimensions;

	fileStream.read( (char*)&numberOfDimensions, sizeof( numberOfDimensions ));

	// **********
	// Dimensions
	// **********
	unsigned long dimensions[ numberOfDimensions ];

	fileStream.read(
		(char *)dimensions,
		numberOfDimensions * sizeof( unsigned long ));

	longVec			size(
					dimensions,
					dimensions + numberOfDimensions );

	// *****************************************
	// Add the array to the end of the data set.
	// *****************************************
	Array *	array = dataSet.addArray( arrayName, dataType, size );

	// ************
	// read Header.
	// ************
	readHeader( fileStream, (Attributable&)*array );

	// *********
	// read Data
	// *********
	fileStream.read(
			(char*)array -> data() -> int8Data(),
			array -> elements() * array -> dataSize() );


	ENDPROC;

}


HiLowDalReaderWriter::HiLowDalReaderWriter()
{
	PROC
	ENDPROC
}



void
HiLowDalReaderWriter::read( DataSet* dataSet )
{

	PROC

	_dataSet = dataSet;

	unsigned int		numberOfBlocks;
	Block::Type		blockType;
	const string 		fileName = _dataSet -> name();
	fstream			fileStream( fileName.c_str(), ios::in );


	// *********************************************************
	// Check to see if the file was opened successfully; if not,
	// then exit.
	// *********************************************************
	if( !fileStream )
	{
		errstr << "could not open file " << fileName << " for reading" << Sas::error("openForReading");

	} /* if */


	// **********************************************************
	// read the comments, history and attributes from the header.
	// **********************************************************
	readHeader(
		fileStream,
		(Attributable&)*_dataSet );


	// **************************
	// read the number of blocks.
	// **************************
	fileStream.read( (char*)&numberOfBlocks, sizeof( numberOfBlocks ));


	// ******************************
	// Do for each block in the file.
	// ******************************
	for(
		unsigned int blockNumber = 0;
		blockNumber < numberOfBlocks;
		blockNumber++ )
	{

		fileStream.read(
				(char*)&blockType,
				sizeof( blockType ));

		switch( blockType )
		{
			case Block::TableT :

				readTable(
					fileStream,
					*_dataSet );

				break;

			case Block::ArrayT :

				readArray(
					fileStream,
					*_dataSet );

				break;

			default :

				errstr << "An invalid block type " << int( blockType ) << " was read from file " << fileName << "; posbbly not a DAL file" << Sas::error("invalidBlockType");

				break;

		} // switch
	} // for

	ENDPROC


}


void
HiLowDalReaderWriter::write( DataSet* dataSet )
{

	PROC

	_dataSet = dataSet;

	unsigned int		numberOfBlocks;
	Block::Type		blockType;
	const string 		fileName = _dataSet -> name();
	fstream			fileStream( fileName.c_str(), ios::out );


	// *********************************************************
	// Check to see if the file was opened successfully; if not,
	// then exit.
	// *********************************************************
	if( !fileStream )
	{
		errstr << "could not open file " << fileName << " for writing" << Sas::error("openForWriting");

	} /* if */

	// ****************************************************
	// write the data set Comments, History and Attributes.
	// ****************************************************
	writeHeader(
		fileStream,
		(const Attributable&)*_dataSet );


	// ***************************
	// write the number of blocks.
	// ***************************
	numberOfBlocks = _dataSet -> blocks();

	fileStream.write(
			(char *)&numberOfBlocks,
			sizeof( numberOfBlocks ));

	// *******************************
	// For each block in the data set.
	// *******************************
	for(
		unsigned int blockNumber = 0;
		blockNumber < numberOfBlocks;
		++blockNumber )
	{

		blockType = _dataSet -> block( blockNumber ) -> type();

		fileStream.write(
			(char *)&blockType,
			sizeof( blockType ));

		switch( blockType )
		{
			case Block::TableT :

				writeTable(
					fileStream,
					*( _dataSet -> table( blockNumber )));

				break;

			case Block::ArrayT :

				writeArray(
					fileStream,
					*( _dataSet -> array( blockNumber )));

				break;

			default :

				errstr << "Could not write block " << _dataSet->block(blockNumber)->name() << " from set " << _dataSet->name() << " because it has an invalid block type " << int( blockType ) << Sas::error("invalidBlockType");

				break;

		} // switch
	} // for

	ENDPROC;

}

