using namespace std;
/*************************************************************** 
* 
* MemDalFilereaderwriter.cc     & XMM Science Analysis System (SAS) 
* 
* Class representing DAL Dataset reader/writer.
* 
* (c) 1997 ESA - Written by Mark Thomas, CARA
* 
* $Id: MemDalReaderWriter.cc,v 1.2 2002/09/06 14:57:38 gvacanti Exp $ 
* 
*****************************************************************/ 


#include <string>			// C++

#include <Tracer.h>
#include <errstr.h>			// SAS
#include "Dal.h"			// DAL

#include "utilities.h"
#include "MemDalReaderWriter.h"
#include "dal_utilities.h"


static void
wrt( fstream & fs, const string& thisString )
{
	PROC

	const int len = thisString.length();
	fs.write( (char*)&len, sizeof( len ) );

	fs.write(
		(char*)thisString.c_str(),
		thisString.length() );

	ENDPROC
}


template<class T>
static void
wrt( fstream & fs, T value )
{
	PROC

	fs.write( (char*)(&value), sizeof( value ) );

	ENDPROC
}


static string
rdstr( fstream & fs )
{
	PROC
	int len;
	fs.read( (char*)&len, sizeof( len ) );
	char nxtstr[ len + 1 ];
	fs.read( nxtstr, len );
	nxtstr[ len ] = '\0';
	RETURN nxtstr;
}


static void
writeComments(
	fstream	&			fileStream,
	const Attributable &		attributeSet )
{

	PROC

	// ***************
	// write Comments.
	// ***************

	// =========================
	// write number of comments.
	// =========================
	wrt( fileStream, attributeSet.comments() );

	// ==========================
	// write each comment in turn.
	// ==========================
	for(
		unsigned int i = 0;
		i < attributeSet.comments();
		++i )
	{
		wrt( fileStream, attributeSet.comment( i ) );

	} // for

	ENDPROC

}


static void
writeHistory(
	fstream	&			fileStream,
	const Attributable &		attributeSet )
{

	PROC

	// ***************
	// write Comments.
	// ***************

	// ==============================
	// write number of history lines.
	// ==============================
	wrt( fileStream, attributeSet.historys() );

	// ================================
	// write each history line in turn.
	// ================================
	for(
		unsigned int i = 0;
		i < attributeSet.historys();
		++i )
	{
		wrt( fileStream, attributeSet.history( i ) );

	} // for

	ENDPROC
}


static void
readComments(
	fstream	&		fileStream,
	Attributable &		attributeSet )
{

	PROC

	unsigned int		numberOfComments;

	// ========================
	// read number of comments.
	// ========================
	fileStream.read(
			(char*)&numberOfComments,
			sizeof( numberOfComments ));

	// ==================================
	// read and set each comment in turn.
	// ==================================
	for( unsigned int i = 0; i < numberOfComments; i++ )
	{

		attributeSet.addComment( rdstr( fileStream ));

	} // for

	ENDPROC

}


static void
readHistory(
	fstream	&			fileStream,
	Attributable &		attributeSet )
{

	PROC

	unsigned int		numberOfHistory;

	// =============================
	// read number of history lines.
	// =============================
	fileStream.read(
			(char*)&numberOfHistory,
			sizeof( numberOfHistory ));

	// ==================================
	// read and set each history in turn.
	// ==================================
	for( unsigned int i = 0; i < numberOfHistory; i++ )
	{

		attributeSet.addHistory( rdstr( fileStream ));

	} // for

	ENDPROC

}


static void
writeAttributes(
	fstream	&			fileStream,
	const Attributable &		attributeSet )
{

	PROC

	// ===========================
	// write number of attributes.
	// ===========================
	wrt( fileStream, attributeSet.attributes() );

	for(
		unsigned int i = 0;
		i < attributeSet.attributes();
		++i )
	{
		const Attribute* attribute = attributeSet.attribute( i );

		// ---------------------
		// write attribute name.
		// ---------------------
		wrt( fileStream, attribute -> name() );

		// ---------------------------
		// write attribute value type.
		// ---------------------------
		wrt( fileStream, attribute -> dataType() );

		// ----------------------
		// write attribute units.
		// ----------------------
		wrt( fileStream, attribute -> units() );

		// ------------------------
		// write attribute comment.
		// ------------------------
		wrt( fileStream, attribute -> label() );

		// ----------------------
		// write attribute value.
		// ----------------------
		switch( attribute -> dataType() )
		{
			case Attribute::Bool :
				wrt( fileStream, attribute -> asBool() );
				break;
			case Attribute::Int :
				wrt( fileStream, attribute -> asInt() );
				break;
			case Attribute::Real :
				wrt( fileStream, attribute -> asReal() );
				break;
			case Attribute::DString :
				wrt( fileStream, attribute -> asString() );
				break;
			default :
				errstr << "Attribute data type " <<
				int(attribute -> dataType()) << " not recognised" << Sas::error( "attributeWriteError" );
		} // switch
	} // for

	ENDPROC

}


static void
readAttributes(
	fstream	&			fileStream,
	Attributable &		attributeSet )
{

	PROC

	unsigned int numberOfAttributes;


	// ****************
	// read Attributes.
	// ****************

	// ==========================
	// read number of attributes.
	// ==========================
	fileStream.read(
			(char*)&numberOfAttributes,
			sizeof( numberOfAttributes ));

	// ====================================
	// read and set each attribute in turn.
	// ====================================
	for( unsigned int i = 0; i < numberOfAttributes; i++ )
	{

		const string attributeName = rdstr( fileStream );
		Attribute::DataType	dataType;
		fileStream.read((char*)&dataType, sizeof( dataType ));
		const string attributeUnits = rdstr( fileStream );
		const string attributeComment = rdstr( fileStream );

		// ==========================
		// Now add the new attribute.
		// ==========================
		Attribute *	attribute =

					attributeSet.addAttribute(
							attributeName,
							attributeComment,
							attributeUnits );

		attribute -> dataType( dataType );

		switch( dataType )
		{
			case Attribute::Bool :
			{
				bool	value;

				fileStream.read( (char*)&value, sizeof(value ));
				*attribute = value;
			}
				break;
			case Attribute::Int :
			{
				int	value;

				fileStream.read( (char*)&value, sizeof(value ));
				*attribute = value;
			}
				break;
			case Attribute::Real :
			{
				double	value;

				fileStream.read( (char*)&value, sizeof(value ));
				*attribute = value;
			}
				break;
			case Attribute::DString :
			{
				*attribute = rdstr( fileStream );
			} 
				break;
			default :
				errstr << "Attribute data type " << int(dataType) << " not recognised" << Sas::error( "attributeReadError" );
		} // switch
	} // for

	ENDPROC

}


static void
writeHeader(
	fstream	&			fileStream,
	const Attributable &		attributeSet )
{

	PROC

	writeAttributes( fileStream, attributeSet );
	writeComments( fileStream, attributeSet  );
	writeHistory( fileStream, attributeSet  );

	ENDPROC;

}


static void
readHeader(
	fstream	&			fileStream,
	Attributable &		attributeSet )
{

	PROC

	readAttributes( fileStream, attributeSet );
	readComments( fileStream, attributeSet );
	readHistory( fileStream, attributeSet );

	ENDPROC;

}


static void
writeVariableLengthColumnData(
	fstream	&		fileStream,
	const Column *		column )
{
	PROC

	for( unsigned int j = 0; j < column -> rows(); ++j )
	{

	 	// *******************************************************
		// Get access to the cell data of the column. This will be
		// deleted after the switch.
 		// *******************************************************
		CellData *data = column->cellData( j );

		wrt( fileStream, data -> size() );

		switch( column -> dataType() )
		{
			case Column::Int8 :
				fileStream.write(
					(char*)data -> int8Data(),
					data -> size() * column -> dataSize() );
				break;

			case Column::Uint16 :
				fileStream.write(
					(char*)data -> uint16Data(),
					data -> size() * column -> dataSize() );
				break;

			case Column::Int16 :
				fileStream.write(
					(char*)data -> int16Data(),
					data -> size() * column -> dataSize() );
				break;

			case Column::Uint32 :
				fileStream.write(
					(char*)data -> uint32Data(),
					data -> size() * column -> dataSize() );
				break;

			case Column::Int32 :
				fileStream.write(
					(char*)data -> int32Data(),
					data -> size() * column -> dataSize() );
				break;

			case Column::Real32 :
				fileStream.write(
					(char*)data -> real32Data(),
					data -> size() * column -> dataSize() );
				break;

			case Column::Real64 :
				fileStream.write(
					(char*)data -> real64Data(),
					data -> size() * column -> dataSize() );
				break;

			case Column::Bool :
				fileStream.write(
					(char*)data -> boolData(),
					data -> size() * column -> dataSize() );
				break;

			case Column::DString :
				fileStream.write(
					(char*)data -> stringData(),
					data -> size() * column -> dataSize() );
				break;

			default :
				errstr << "Invalid column type" << Sas::error( "invalidColumn" );
				break;

		} // switch

		// Delete the data object again.
		delete data;

	} // for

	ENDPROC
}


static void
writeFixedLengthColumnData(
	fstream	&		fileStream,
	const Column *		column )
{
	PROC

 	// ***********************************************************
	// Get access to the data of the column. This is to be deleted
        // after the switch again.
 	// ***********************************************************
	ColumnData *data = column->data();

	switch( column -> dataType() )
	{
		case Column::Int8 :
			fileStream.write(
				(char*)data -> int8Data(),
				column -> rows() * column -> elements() * column -> dataSize() );
			break;
		case Column::Uint16 :
			fileStream.write(
				(char*)data -> uint16Data(),
				column -> rows() * column -> elements() * column -> dataSize() );
			break;
		case Column::Int16 :
			fileStream.write(
				(char*)data -> int16Data(),
				column -> rows() * column -> elements() * column -> dataSize() );
			break;
		case Column::Uint32 :
			fileStream.write(
				(char*)data -> uint32Data(),
				column -> rows() * column -> elements() * column -> dataSize() );
			break;
		case Column::Int32 :
			fileStream.write(
				(char*)data -> int32Data(),
				column -> rows() * column -> elements() * column -> dataSize() );
			break;
		case Column::Real32 :
			fileStream.write(
				(char*)data -> real32Data(),
				column -> rows() * column -> elements() * column -> dataSize() );
			break;
		case Column::Real64 :
			fileStream.write(
				(char*)data -> real64Data(),
				column -> rows() * column -> elements() * column -> dataSize() );
			break;
		case Column::Bool :
			fileStream.write(
				(char*)data -> boolData(),
				column -> rows() * column -> elements() * column -> dataSize() );
			break;
		case Column::DString :
			fileStream.write(
				(char*)data -> stringData(),
				column -> rows() * column -> elements() * column -> dataSize() );
			break;
		default :
			errstr << "Invalid column type" << Sas::error( "invalidColumn" );
			break;
	} // switch

	// Delete the data object again.
	delete data;

	ENDPROC
}


static void
writeColumnNullValue(
	fstream	&		fileStream,
	const Column*		column )
{
	PROC

	wrt(
		fileStream,
		column -> nullDefined() && column -> nullType() == Nullable::Integer );

	if( column -> nullDefined() && column -> nullType() == Nullable::Integer )
	{
		wrt( fileStream, column -> intNullValue() );
	} // if
	ENDPROC
}


static void
writeColumns(
	fstream	&		fileStream,
	const Table* 		table )
{
	PROC

	wrt( fileStream, table -> columns() );

	// ********************************
	// Do for each column in the table:
	//   name
	//   data type
	//   units
	//   number of dimensions
	//   dimensions
	//   data
	// ********************************
	for( unsigned int ic = 0; ic < table -> columns(); ic++ )
	{
		const Column *	column = table -> column( ic );

		// ====
		// name
		// ====
		wrt( fileStream, column -> name() );

		// =========
		// data type
		// =========
		wrt( fileStream, column -> dataType() );

		// =====
		// Units
		// =====
		wrt( fileStream, column -> units() );

		// =======
		// Comment
		// =======
		wrt( fileStream, column -> label() );

		// ==============================================
		// Number of dimensions. The number of dimensions
		// may be zero.
		// ==============================================
		wrt( fileStream, column -> dimensions() );

		// ==========
		// Dimensions
		// ==========
		fileStream.write(
			(char *)Mvec<unsigned long>( column -> size() ).data(),
			column -> dimensions() * sizeof( unsigned long ) );

		// ========================
		// Write column null value.
		// ========================
		writeColumnNullValue( fileStream, column );

		// ==============
		// Fixed/Variable
		// ==============
		wrt( fileStream, column -> cellType() );

		// ====
		// Data
		// ====
		if( column -> cellType() == Column::Fixed )
			writeFixedLengthColumnData( fileStream, column );
		else	
			writeVariableLengthColumnData( fileStream, column );

		// ========================
		// Write Column Attributes.
		// ========================
		writeHeader( fileStream, *column );
	} // for

	ENDPROC
}


static void
writeTable(
	fstream	&		fileStream,
	const Table* 		table )
{

	PROC


	// ********************
	// write the Table name.
	// ********************
	wrt( fileStream, table -> name() );

	// **************************************
	// write the number of rows in the table.
	// **************************************
	wrt( fileStream, table -> rows() );

	// *************
	// write Header.
	// *************
	writeHeader( fileStream, *table );

	// ************************
	// Write the table columns.
	// ************************
	writeColumns( fileStream, table );

	ENDPROC

}


static void
readColumnNullValue(
	fstream	&	fileStream,
	Column*		column )
{
	PROC

	bool		nullDefined;

	fileStream.read( (char*)&nullDefined, sizeof( nullDefined ) );

	if( nullDefined )
	{
		int	nullValue;
		fileStream.read( (char*)&nullValue, sizeof( nullValue ));
		column -> nullValue( nullValue );
	} // if

	ENDPROC
}


static void
readFixedLengthColumnData(
	fstream	&	fileStream,
	Column*		column )
{
	PROC

	ColumnData*	data = column -> data();

	switch( column -> dataType() )
	{
		case Column::Int8 :
			fileStream.read(
				(char*)data -> int8Data(),
				column -> rows() * column -> dataSize() * column -> elements() );
			break;
		case Column::Uint16 :
			fileStream.read(
				(char*)data -> uint16Data(),
				column -> rows() * column -> dataSize() * column -> elements() );
			break;
		case Column::Int16 :
			fileStream.read(
				(char*)data -> int16Data(),
				column -> rows() * column -> dataSize() * column -> elements() );
			break;
		case Column::Uint32 :
			fileStream.read(
				(char*)data -> uint32Data(),
				column -> rows() * column -> dataSize() * column -> elements() );
			break;
		case Column::Int32 :
			fileStream.read(
				(char*)data -> int32Data(),
				column -> rows() * column -> dataSize() * column -> elements() );
			break;
		case Column::Real32 :
			fileStream.read(
				(char*)data -> real32Data(),
				column -> rows() * column -> dataSize() * column -> elements() );
			break;
		case Column::Real64 :
			fileStream.read(
				(char*)data -> real64Data(),
				column -> rows() * column -> dataSize() * column -> elements() );
			break;
		case Column::Bool :
			fileStream.read(
				(char*)data -> boolData(),
				column -> rows() * column -> dataSize() * column -> elements() );
			break;
		case Column::DString :
			fileStream.read(
				(char*)data -> stringData(),
				column -> rows() * column -> dataSize() * column -> elements() );
			break;
		default :
			errstr << "Invalid column type " << Sas::error( "columnReadError" );
	} // switch

	delete data;

	ENDPROC
}



static void
readVariableLengthColumnData(
	fstream	&	fileStream,
	Column*		column )
{
	PROC

	for( unsigned int i = 0; i < column -> rows(); ++i )
	{
		CellData*	data = column -> cellData( i );

		switch( column -> dataType() )
		{
			case Column::Int8 :
				fileStream.read(
					(char*)data -> int8Data(),
					column -> dataSize() * data -> size() );
				break;
			case Column::Uint16 :
				fileStream.read(
					(char*)data -> uint16Data(),
					column -> dataSize() * data -> size() );
				break;
			case Column::Int16 :
				fileStream.read(
					(char*)data -> int16Data(),
					column -> dataSize() * data -> size() );
				break;
			case Column::Uint32 :
				fileStream.read(
					(char*)data -> uint32Data(),
					column -> dataSize() * data -> size() );
				break;
			case Column::Int32 :
				fileStream.read(
					(char*)data -> int32Data(),
					column -> dataSize() * data -> size() );
				break;
			case Column::Real32 :
				fileStream.read(
					(char*)data -> real32Data(),
					column -> dataSize() * data -> size() );
				break;
			case Column::Real64 :
				fileStream.read(
					(char*)data -> real64Data(),
					column -> dataSize() * data -> size() );
				break;
			case Column::Bool :
				fileStream.read(
					(char*)data -> boolData(),
					column -> dataSize() * data -> size() );
				break;
			case Column::DString :
				fileStream.read(
					(char*)data -> stringData(),
					column -> dataSize() * data -> size() );
				break;
			default :
				errstr << "Invalid column type " << Sas::error( "columnReadError" );
		} // switch

		delete data;

	} // for

	ENDPROC
}



static void
readColumn(
	fstream	&	fileStream,
	Table*		table )
{
	PROC

	Column*		column = 0;

	// ====
	// name
	// ====
	const string	columnName( rdstr( fileStream ));

	// =========
	// data type
	// =========
	Dal::DataType		dataType;
	fileStream.read( (char*)&dataType, sizeof( dataType ));

	// =====
	// units
	// =====
	const string	units( rdstr( fileStream ));

	// =======
	// comment
	// =======
	const string	comment( rdstr( fileStream ));

	// ====================
	// number of dimensions
	// ====================
	unsigned int		numberOfDimensions;
	fileStream.read(
			(char*)&numberOfDimensions,
			sizeof( numberOfDimensions ));

	// ==========
	// dimensions
	// ==========
	unsigned long	dimensions[ numberOfDimensions ];

	fileStream.read(
		(char *)dimensions,
		numberOfDimensions * sizeof( unsigned long ));

	readColumnNullValue( fileStream, column );

	Column::CellType	cellType;

	fileStream.read( (char*)&cellType, sizeof( cellType ));

	// ============================
	// Add the column to the table.
	// ============================
	if( cellType == Column::Fixed )
	{
		vector<unsigned long>	size(
						dimensions,
						dimensions + numberOfDimensions );

		column = table -> addColumn(
						columnName,
						dataType,
						comment,
						units,
						size );
	}
	else
	{
		column = table -> addColumn(
					columnName,
					dataType,
					comment,
					units,
					Dal::zero );
	} // if

	// =====
	// data.
	// =====
	if( column -> cellType() == Column::Fixed )
	{
		readFixedLengthColumnData( fileStream, column );
	}
	else
	{
		readVariableLengthColumnData( fileStream, column );
	} // if

	// ***********************
	// Read Column Attributes.
	// ***********************
	readHeader( fileStream, *column );

	ENDPROC
}


static void
readTable(
	fstream	&	fileStream,
	DataSet &	dataSet )
{
	PROC

	unsigned int		numberOfColumns;
	unsigned long		numberOfRows;


	// ****************************************************
	// Do:
	//	The table Name - read
	//	The nunber of rows in the table columns
	// ****************************************************
	const string	tableName( rdstr( fileStream ));

	fileStream.read(
			(char*)&numberOfRows,
			sizeof( numberOfRows ));

	// *****************************************
	// Add the table to the end of the data set.
	// *****************************************
	Table * table = dataSet.addTable( tableName, numberOfRows );

	// ************
	// read Header.
	// ************
	readHeader( fileStream, *table );


	// ***********************
	// read number of columns.
	// ***********************
	fileStream.read(
			(char*)&numberOfColumns,
			sizeof( numberOfColumns ));



	// ********************************
	// Do for each column in the table:
	//   name
	//   datatype
	//   units
	//   number of dimensions
	//   dimensions
	//   data
	// ********************************
	for( unsigned int ic = 0; ic < numberOfColumns; ic++ )
	{
		readColumn( fileStream, table );
	} // for

	ENDPROC

}


static void
writeArrayData(
	fstream &	fileStream,
	const Array*	array )
{
	PROC

	// **********************************
	// Exit if there is no data to write.
	// **********************************
	if( numberOfElements( array -> size() ) == 0 )
	{
		ALTRETURN;
	} // if

	ArrayData*	data = array -> data();

	switch( array -> dataType() )
	{
		case Array::Int8 :
			fileStream.write(
				(char*)data -> int8Data(),
				array -> elements() * array -> dataSize() );
			break;
		case Array::Uint16 :
			fileStream.write(
				(char*)data -> uint16Data(),
				array -> elements() * array -> dataSize() );
			break;
		case Array::Int16 :
			fileStream.write(
				(char*)data -> int16Data(),
				array -> elements() * array -> dataSize() );
			break;
		case Array::Uint32 :
			fileStream.write(
				(char*)data -> uint32Data(),
				array -> elements() * array -> dataSize() );
			break;
		case Array::Int32 :
			fileStream.write(
				(char*)data -> int32Data(),
				array -> elements() * array -> dataSize() );
			break;
		case Array::Real32 :
			fileStream.write(
				(char*)data -> real32Data(),
				array -> elements() * array -> dataSize() );
			break;
		case Array::Real64 :
			fileStream.write(
				(char*)data -> real64Data(),
				array -> elements() * array -> dataSize() );
			break;
		case Array::Bool :
			fileStream.write(
				(char*)data -> boolData(),
				array -> elements() * array -> dataSize() );
			break;
		case Array::DString :
			fileStream.write(
				(char*)data -> stringData(),
				array -> elements() * array -> dataSize() );
			break;
		default :
			errstr << "Invalid array type " << Sas::error( "arrayWriteError" );
	} // switch

	delete data;

	ENDPROC
}


static void
writeArray(
	fstream &	fileStream,
	const Array*	array )
{
	PROC

	// *********************
	// write the Array name.
	// *********************
	wrt( fileStream, array -> name() );


	// ****************************
	// write:
	//	Data type.
	//	Number of dimensions
	//	Size of each dimension
	//	Data
	// ****************************

	// =========
	// Data Type
	// =========
	wrt( fileStream, array -> dataType() );

	// ====================
	// Number of dimensions
	// ====================
	wrt( fileStream, array -> dimensions() );

	// ==========
	// Dimensions
	// ==========
	fileStream.write(
		(char *)Mvec<unsigned long>( array -> size() ).data(),
		array -> dimensions() * sizeof( unsigned long ) );

	// ====
	// Data
	// ====
	writeArrayData( fileStream, array );

	// *************
	// write Header.
	// *************
	writeHeader( fileStream, *array );

	ENDPROC

}


static void
readArrayData(
	fstream &	fileStream,
	Array*		array )
{
	PROC

	// *********************************
	// Exit if there is no data to read. 
	// *********************************
	if( numberOfElements( array -> size() ) == 0 )
	{
		ALTRETURN;
	} // if

	ArrayData*	data = array -> data();

	switch( array -> dataType() )
	{
		case Array::Int8 :
			fileStream.read(
				(char*)data -> int8Data(),
				array -> elements() * array -> dataSize() );
			break;
		case Array::Uint16 :
			fileStream.read(
				(char*)data -> uint16Data(),
				array -> elements() * array -> dataSize() );
			break;
		case Array::Int16 :
			fileStream.read(
				(char*)data -> int16Data(),
				array -> elements() * array -> dataSize() );
			break;
		case Array::Uint32 :
			fileStream.read(
				(char*)data -> uint32Data(),
				array -> elements() * array -> dataSize() );
			break;
		case Array::Int32 :
			fileStream.read(
				(char*)data -> int32Data(),
				array -> elements() * array -> dataSize() );
			break;
		case Array::Real32 :
			fileStream.read(
				(char*)data -> real32Data(),
				array -> elements() * array -> dataSize() );
			break;
		case Array::Real64 :
			fileStream.read(
				(char*)data -> real64Data(),
				array -> elements() * array -> dataSize() );
			break;
		case Array::Bool :
			fileStream.read(
				(char*)data -> boolData(),
				array -> elements() * array -> dataSize() );
			break;
		case Array::DString :
			fileStream.read(
				(char*)data -> stringData(),
				array -> elements() * array -> dataSize() );
			break;
		default :
			errstr << "Invalid array type " << Sas::error( "arrayReadError" );

	} // switch

	delete data;

	ENDPROC
}


static void
readArray(
	fstream &	fileStream,
	DataSet &	dataSet )
{

	PROC


	// ********************
	// read the array name.
	// ********************
	const string arrayName = rdstr( fileStream );

	// *********
	// data type
	// *********
	Array::DataType		dataType;

	fileStream.read( (char*)&dataType, sizeof( dataType ));

	// ********************
	// Number of dimensions
	// ********************
	unsigned int		numberOfDimensions;

	fileStream.read( (char*)&numberOfDimensions, sizeof( numberOfDimensions ));

	// **********
	// Dimensions
	// **********
	unsigned long dimensions[ numberOfDimensions ];

	fileStream.read(
		(char *)dimensions,
		numberOfDimensions * sizeof( unsigned long ));

	vector<unsigned long>	size(
					dimensions,
					dimensions + numberOfDimensions );

	// *****************************************
	// Add the array to the end of the data set.
	// *****************************************
	Array *	array = dataSet.addArray( arrayName, dataType, size );

	// *********
	// read Data
	// *********
	readArrayData( fileStream, array );

	// ************
	// read Header.
	// ************
	readHeader( fileStream, *array );

	ENDPROC

}


MemDalReaderWriter::MemDalReaderWriter()
{
	PROC
	ENDPROC
}



void
MemDalReaderWriter::read( DataSet* dataSet )
{

	PROC

	_dataSet = dataSet;

	unsigned int		numberOfBlocks;
	Block::Type		blockType;

	const string		fileName = _dataSet -> name();

	// ******************************
	// Check that the dataset exists.
	// ******************************
	if( !dataSetServer -> exists( fileName ) )
	{
		errstr << "The dataset with name '" << fileName << "' could not be read; either it does not exist or it is in an unrecognised format" << Sas::error( "datasetCouldNotBeRead" );
	} // if

	// **********************************************
	// Open the Dal Data File.
	// Then check that the file opened without error.
	// **********************************************
	_fs.open( fileName.c_str(), ios::in );

	if( !_fs )
	{
		errstr << "could not open file " << fileName << " for reading" << Sas::error("openForReading");

	} /* if */

	// **********************************************************
        // Check magic number
        // **********************************************************
	int n = magic().length();
	char m[n+1];
	_fs.read(m,n);
	m[n] = 0;
	if ( ! (magic() == string(m)) )
	{
		errstr << "file " << fileName << " is not in DAL file format" << Sas::error("notDalFormat");
	} // if

	// **********************************************************
	// read the comments, history and attributes from the header.
	// **********************************************************
	readHeader( _fs, *_dataSet );


	// **************************
	// read the number of blocks.
	// **************************
	_fs.read( (char*)&numberOfBlocks, sizeof( numberOfBlocks ));


	// ******************************
	// Do for each block in the file.
	// ******************************
	for(
		unsigned int blockNumber = 0;
		blockNumber < numberOfBlocks;
		blockNumber++ )
	{

		_fs.read( (char*)&blockType, sizeof( blockType ));

		switch( blockType )
		{
			case Block::TableT :

				readTable(
					_fs,
					*_dataSet );

				break;

			case Block::ArrayT :

				readArray(
					_fs,
					*_dataSet );

				break;

			default :

				errstr << "An invalid block type " << int( blockType ) << " was read from file " << fileName << "; posbbly not a DAL file" << Sas::error("invalidBlockType");

				break;

		} // switch
	} // for

	_fs.close();

	ENDPROC

}


void
MemDalReaderWriter::write( DataSet* dataSet )
{

	PROC

	_dataSet = dataSet;

	const string		setName = dataSet -> name();

	dataSetServer -> clobberDataSet( setName );

	// ******************************
	// Attempt to create the dataset.
	// ******************************
	_fs.open( setName.c_str(), ios::out );

	if( !_fs )
	{
		errstr << "The dataset with name '" << setName << "' could not be created." << Sas::error("createDatasetFailed");

	} // if


	// ****************************************************
        // write magic number
        // ****************************************************
	_fs.write( magic().c_str(), magic().length() );

	// ****************************************************
	// write the data set Comments, History and Attributes.
	// ****************************************************
	writeHeader( _fs, *_dataSet );

	// ***************************
	// write the number of blocks.
	// ***************************
	const unsigned int	numberOfBlocks = dataSet -> blocks();

	wrt( _fs, numberOfBlocks );

	// *******************************
	// For each block in the data set.
	// *******************************
	for(
		unsigned int blockNumber = 0;
		blockNumber < numberOfBlocks;
		++blockNumber )
	{
		const Block::Type blockType = _dataSet -> block( blockNumber ) -> type();

		wrt( _fs, blockType );

		switch( blockType )
		{
			case Block::TableT :

				writeTable(
					_fs,
					_dataSet -> table( blockNumber ));

				break;

			case Block::ArrayT :

				writeArray(
					_fs,
					_dataSet -> array( blockNumber ));

				break;

			default :

				errstr << "Could not write block '" << _dataSet->block(blockNumber)->name() << "' from set '" << _dataSet->name() << "' because it has an invalid block type " << int( blockType ) << Sas::error("invalidBlockType");

				break;

		} // switch
	} // for

	_fs.close();

	ENDPROC

}


